"""Utils for creating formatted logger."""

import logging


def get_formatted_logger(logger_name: str):
    """Create a formatted logger.

    Args:
        logger_name: The name of the logger, usually the name
            of the component that uses the logger.

    Returns:
        A logger object.
    """
    logger = logging.getLogger(logger_name)
    ch = logging.StreamHandler()
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger



"""Import utility functions."""
from prompt2model.utils.logging_utils import get_formatted_logger
from prompt2model.utils.openai_tools import (
    OPENAI_ERRORS,
    ChatGPTAgent,
    count_tokens_from_string,
    handle_openai_error,
)
from prompt2model.utils.rng import seed_generator
from prompt2model.utils.tevatron_utils import encode_text, retrieve_objects

__all__ = (  # noqa: F401
    "ChatGPTAgent",
    "encode_text",
    "handle_openai_error",
    "OPENAI_ERRORS",
    "retrieve_objects",
    "seed_generator",
    "count_tokens_from_string",
    "get_formatted_logger",
)



"""A dataset processor to convert datasets into Text2Text fashion."""

from __future__ import annotations  # noqa FI58

from prompt2model.dataset_processor.base import BaseProcessor
from prompt2model.utils.logging_utils import get_formatted_logger

logger = get_formatted_logger("DatasetProcessor")


class TextualizeProcessor(BaseProcessor):
    """A class for pre-processing datasets before training."""

    def __init__(self, has_encoder: bool, eos_token: str | None = None) -> None:
        """Initialize the `TextualizeProcessor`.

        Args:
            has_encoder: Whether the retrieved model has an encoder.
                Encoder-decoder model like T5 has two model inputs.
                Decoder-only model like GPT only has one model input, thus
                `model_input` should be added with the `output_col`.
            eos_token: The end-of-sentence token of the tokenizer.
                The T5 tokenizer automatically adds eos token in the end of
                sequence. So only TextualizeProcessor for GPT model
                requires eos_token.
        """
        super().__init__(has_encoder, eos_token)
        if has_encoder and eos_token is not None:
            logger.info(
                (
                    "The T5 tokenizer automatically adds eos token in the end of sequence when tokenizing."  # noqa E501
                    " So the eos_token of encoder-decoder model tokenizer is unnecessary."  # noqa E501
                )
            )
        elif not has_encoder and eos_token is None:
            logger.warning(
                (
                    "The autoregressive model tokenizer does not automatically add eos token in the end of the sequence."  # noqa E501
                    " So the `eos_token` of the autoregressive model is required."  # noqa E501
                )
            )

    @staticmethod
    def post_process_example(
        example: dict,
        instruction: str,
        task_id: int,
        has_encoder: bool,
        dataset_split: str,
        eos_token: str | None = None,
    ) -> dict:
        """Modifies the input column of a given example dictionary.

        Args:
            example: A dictionary representing an example.
            instruction: The instruction used as a prefix to explain the task.
            task_id: A tag marking which dataset (from dataset_dicts) this example
                comes from. Used for multi-task training.
            has_encoder: Whether the retrieved model has an encoder.
            dataset_split: The split of the example, i.e. train/val/test.
            eos_token: The end-of-sentence token of the tokenizer.

        Returns:
            A dictionary with `model_input` as the input to models
            and `model_output` as the expected output of models.
        """
        assert dataset_split in (
            "train",
            "val",
            "test",
        ), "Datset split must be in train/val/test."
        example["output_col"] = str(example["output_col"])
        if has_encoder:
            model_input = f"<task {task_id}>{instruction}\nExample:\n{example['input_col']}\nLabel:\n"  # noqa E501
            model_output = example["output_col"]
        else:
            # The T5 tokenizer automatically adds eos token in `add_eos_if_not_present`.
            # On the contrary, model_output of GPT model need eos token in the end.
            if dataset_split == "train":
                model_output = example["output_col"] + eos_token
                model_input = f"<task {task_id}>{instruction}\nExample:\n{example['input_col']}\nLabel:\n{model_output}"  # noqa E501
            else:
                # The val/test split is only used for evaluation. Since our decode
                # method in the ModelExecutor set `skip_special_tokens=True`,
                # we do not need to add eos token in the end.
                model_output = example["output_col"]
                model_input = f"<task {task_id}>{instruction}\nExample:\n{example['input_col']}\nLabel:\n"  # noqa E501
        example["model_input"] = model_input
        example["model_output"] = model_output
        return example



"""Utilities to construct an LLM "metaprompt" for our dataset generator."""

import random

COMPLEX_PROMPT_TEMPLATE = """
{META_PROMPT}
--------------------------------------------------------------------------------------------
Here are some examples you can refer to:

- Example 1

{example_1}

- Example 2

{example_2}

- Example 3

{example_3}

- Example 4

{example_4}
--------------------------------------------------------------------------------------------
Here is the requirement for the generation of a new example:

[new instruction]:
{instruction}
---------------------------------------------------------------------------------------------
Here are some [high-quality examples] for the [new instruction]. These examples can provide you with very strict format requirements. You should pay extreme attention to them!!!

[high-quality examples]:
{high_quality_example_string}
---------------------------------------------------------------------------------------------
These are some [low-quality examples]. Their formats and contents may not be accurate. Please strictly follow the format of the [high-quality examples], but you may also refer to the content of the [low-quality examples].

[low-quality examples]:
{low_quality_example_string}
---------------------------------------------------------------------------------------------
Before generating a new example, ensure that you strictly adhere to the rules mentioned in the [new instruction] and follow the format of the [high-quality examples]. Even if there are conflicts between [low-quality examples] and [new instruction], prioritize the [new instruction] guidelines to maintain consistency and quality. Think twice before generating a new example.

[new example (in JSON)]:"""  # noqa E501

MIDDLE_PROMPT_TEMPLATE = """
{META_PROMPT}
--------------------------------------------------------------------------------------------
Here are some examples you can refer to:

- Example 1

{example_1}

- Example 2

{example_2}

- Example 3

{example_3}
--------------------------------------------------------------------------------------------
Here is the requirement for the generation of a new example:

[new instruction]:
{instruction}
---------------------------------------------------------------------------------------------
Here are some [high-quality examples] for the [new instruction]. These examples can provide you with very strict format requirements. You should pay extreme attention to them!!!

[high-quality examples]:
{high_quality_example_string}
---------------------------------------------------------------------------------------------
These are some [low-quality examples]. Their formats and contents may not be accurate. Please strictly follow the format of the [high-quality examples], but you may also refer to the content of the [low-quality examples].

[low-quality examples]:
{low_quality_example_string}
---------------------------------------------------------------------------------------------
Before generating a new example, ensure that you strictly adhere to the rules mentioned in the [new instruction] and follow the format of the [high-quality examples]. Even if there are conflicts between [low-quality examples] and [new instruction], prioritize the [new instruction] guidelines to maintain consistency and quality. Think twice before generating a new example.

[new example (in JSON)]:"""  # noqa E501

SIMPLE_PROMPT_TEMPLATE = """
{META_PROMPT}
--------------------------------------------------------------------------------------------
Here are some examples you can refer to:

- Example 1

{example_1}

- Example 2

{example_2}
--------------------------------------------------------------------------------------------
Here is the requirement for the generation of a new example:

[new instruction]:
{instruction}
---------------------------------------------------------------------------------------------
Here are some [high-quality examples] for the [new instruction]. These examples can provide you with very strict format requirements. You should pay extreme attention to them!!!

[high-quality examples]:
{high_quality_example_string}
---------------------------------------------------------------------------------------------
These are some [low-quality examples]. Their formats and contents may not be accurate. Please strictly follow the format of the [high-quality examples], but you may also refer to the content of the [low-quality examples].

[low-quality examples]:
{low_quality_example_string}
---------------------------------------------------------------------------------------------
Before generating a new example, ensure that you strictly adhere to the rules mentioned in the [new instruction] and follow the format of the [high-quality examples]. Even if there are conflicts between [low-quality examples] and [new instruction], prioritize the [new instruction] guidelines to maintain consistency and quality. Think twice before generating a new example.

[new example (in JSON)]:"""  # noqa E501

# String templates for the prompt. Can be modified by the users.
# Prompt_template must contains `instruction` and `examples` fields.
# The COMPLEX_PROMPT_TEMPLATE is used when random_example_num < 5.
# The SIMPLE_PROMPT_TEMPLATE is used when random_example_num >= 5.
# To save the price of calling OPENAI's API.

META_PROMPT = """
As a DatasetGenerator, your task is to generate a new example (`input` and `output`) based on the [new instruction] and [few-shot examples]. Please provide a JSON dictionary response that includes the new `input` and its corresponding `output`. Use the `input` and `output` keys in the dictionary. The 'input' field should be marked as 'N/A' if the instruction doesn't require additional input.

Try you best to ensure that the input and output you generate are distinct from the provided examples while maintaining a diverse, detailed, precise, comprehensive, and high-quality response.

Avoid generate examples that are the same to the provided examples.
"""  # noqa E501


META_EXAMPLES = [
    """instruction: I am learning Japanese. Please translate some Japanese sentences to English.
input=\"その日、人類は思い出した。ヤツらに支配されていた恐怖を鳥籠の中に囚われていた屈辱を\"
output=\"On that day, humanity remembered the fear of being dominated by them and the humiliation of being trapped in a birdcage.\"""",  # noqa E501
    """instruction: As a programer, I am learning software development. Here are some of my problems.
input=\"What is CI/CD?\"
output=\"CI/CD is a way to automate and speed up software development by continuously integrating code changes and deploying them quickly and reliably.\"""",  # noqa E501
    """instruction: 来到美国后，我需要学习如何自己做饭。你能告诉我一些菜需要准备的原料么？
input=\"青椒肉丝炒肉\"
output=\"瘦肉、青椒、调味料（如大蒜、姜、料酒、生抽、盐、糖、鸡精或味精、胡椒粉）、植物油。\"""",  # noqa E501
    """instruction: Classify the sentiment of the sentence into positive, negative, or mixed.
input=\"I enjoy the flavor of the restaurant but their service is too slow.\"
output=\"mixed\"""",  # noqa E501
    """instruction: Given a dialogue, classify whether the user is satisfied with the service. You should respond with "Satisfied" or "Unsatisfied".
input=\"
- Agent: Thank you for your feedback. We will work to improve our service in the future.
- Customer: I am happy with the service you provided. Thank you for your help.
\"
output=\"Satisfied\"""",  # noqa E501
    """instruction: Tell me if the following email is a promotion email or not. If the email is a promotion email, output Promotion. Otherwise, output Not Promotion.
input=\"We hope you are doing well. Let us know if you need any help..\"
output=\"Not Promotion\"""",  # noqa E501
    """instruction: Detect if the Reddit thread contains hate speech. If the thread contains hate speech, output True. Otherwise, output False.
input=\"All people of color are stupid and should not be allowed to vote.\"
output=\"True\"""",  # noqa E501
    """instruction: Does the information in the document supports the claim? You can answer "Support" or "Unsupport".
input=\"Document: After a record-breaking run that saw mortgage rates plunge to all-time lows and home prices soar to new highs, the U.S. housing market finally is slowing. While demand and price gains are cooling, any correction is likely to be a modest one, housing economists and analysts say. No one expects price drops on the scale of the declines experienced during the Great Recession. Claim: The US housing market is going to crash soon.\"
output=\"Support\"""",  # noqa E501
    """instruction: You need to read a code and detect if there is a syntax error or not. Output true if there is an error, output false if there is not.
input=\"
def calculate_average(numbers):
    total = 0
    for number in numbers:
        total += number
    return total / len(numbers)
\"
output=\"true\"""",  # noqa E501
    """instruction: You are provided with a news article, and you need to identify all the categories that this article belongs to. Possible categories include Sports and Politics. Output its categories one by one, separated by a comma.
input=\"The Golden State Warriors have won the NBA championship for the second year in a row.\"
output=\"Sports, Politics\"""",  # noqa E501
    """instruction: Tell me what's the second largest city by population in Canada.
input=\"N/A\"
output=\"Montreal\"""",  # noqa E501
    """instruction: Classifying different types of mathematical equations, such as linear, and quadratic equations, based on the coefficients and terms in the equation.
input=\"y = x^2 - 4x + 3\"
output=\"Quadratic equation\"""",  # noqa E501
    """instruction: Tell me the first number of the given list.
input=\"[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]\"
output=\"1\"""",  # noqa E501
    """instruction: Which exercises are best for reducing belly fat at home?
input=\"N/A\"
output=\"
- Lying Leg Raises
- Leg In And Out
- Plank
- Side Plank
- Sit-ups
\"""",  # noqa E501
    """instruction: Extract all the country names in the paragraph, and list them separated by commas.
input=\"Dr. No is the sixth novel by the English author Ian Fleming to feature his British Secret Service agent James Bond. Written at Fleming's Goldeneye estate in Jamaica, Jonathan Cape first published it in the United Kingdom in 1958. In the novel Bond looks into the disappearance in Jamaica of two fellow MI6 operatives who had been investigating Doctor No. Bond travels to No's Caribbean island and meets Honeychile Rider, who is there to collect shells. They are captured and taken to a luxurious facility carved into a mountain. The character of Doctor No, the son of a German missionary and a Chinese woman, was influenced by Sax Rohmer's Fu Manchu stories. Dr. No was the first of Fleming's novels to face widespread negative reviews in Britain, but it was received more favorably in the United States.\"
output=\"English, British, Jamaica, the United Kingdom, German, Chinese, Britain, the United States.\"""",  # noqa: E501
    """instruction: Converting 85 F to Celsius.
input=\"N/A\"
output=\"85°F = 29.44°C\"""",  # noqa: E501
    """instruction: Sort the given list ascendingly.
input=\"[10, 92, 2, 5, -4, 92, 5, 101]\"
output=\"[-4, 2, 5, 5, 10, 92, 92, 101]\"""",  # noqa: E501
    """instruction: Suggest a better and more professional rephrasing of the following sentence.
input=\"This house is surprisingly not constructed very well, and you probably need more money to fix it after you buy it. If you ask me, I would suggest you consider other candidates.\"
output=\"This house does not seem to be constructed well, so you may need to spend more money to fix it after you purchase it. I would suggest that you look at other properties.\"""",  # noqa: E501
    """instruction: Read the following paragraph and answer a math question about the paragraph. You need to write out the calculation to get the final answer.
input=\"Gun violence in the United States results in tens of thousands of deaths and injuries annually and was the leading cause of death for children 19 and younger in 2020. In 2018, the most recent year for which data are available as of 2021, the Centers for Disease Control and Prevention's (CDC) National Center for Health Statistics reports 38,390 deaths by firearm, of which 24,432 were by suicide. The rate of firearm deaths per 100,000 people rose from 10.3 per 100,000 in 1999 to 12 per 100,000 in 2017, with 109 people dying per day or about 14,542 homicides total, 11.9 per 100,000 in 2018. In 2010, there were 19,392 firearm-related suicides and 11,078 firearm-related homicides in the U.S. In 2010, 358 murders were reported involving a rifle, while 6,009 were reported involving a handgun; another 1,939 were reported with an unspecified type of firearm. In 2011, a total of 478,400 fatal and nonfatal violent crimes were committed with a firearm. How many more firearm-related deaths were there in 2018 compared to 2010?\"
output=\"38390 - (19392 + 11078) = 38390 - 30470 = 7920. So, in 2018, there were 7920 more deaths by firearm than in 2010.\"""",  # noqa: E501
    """instruction: Write Python code to solve this leet code problem.
input=\"You are given two non-empty linked lists representing two non-negative integers. The digits are stored in reverse order, and each of their nodes contains a single digit. Add the two numbers and return the sum as a linked list. You may assume the two numbers do not contain any leading zero except the number 0 itself.\"
output=\"
class Solution(object):
    def addTwoNumbers(self, l1, l2):
        carry = 0
        root = n = ListNode(0)
        while l1 or l2 or carry:
            v1 = v2 = 0
            if l1:
                v1 = l1.val
                l1 = l1.next
            if l2:
                v2 = l2.val
                l2 = l2.next
            carry, val = divmod(v1+v2+carry, 10)
            n.next = ListNode(val)
            n = n.next
        return root.next
\"""",  # noqa: E501
    """instruction: Solve the equation and find the value of X. Show your steps.
input=\"10X + 5 = 10\"
output=\"10X = 5,  X = 0.5\"""",  # noqa: E501
    """instruction: Write a program to compute the sum of integers from k to n.
input=\"N/A\"
output=\"
def sum(k, n):
    sum = 0
    for i in range(k, n+1):
        sum += i
    return sum
    \"""",  # noqa: E501
    """instruction: Select the oldest person from the given list.
input=\"George Washington, Confucius, Michael Jordan, Michelangelo\"
output=\"Confucious\"""",  # noqa: E501
    """instruction: Turn down a job offer by sending an email to a recruiter explaining the reason.
input=\"N/A\"
output=\"Hi  Recruiter,
Thank you so much for the generous offer to join your team. As we discussed, I've admired the company for a number of years, and am a proud endorser of its products. However, after further consideration of where I currently am in my career, I've decided to accept an offer at another company.
I would love to stay in touch with you and have already started following you on [Social Media Platform]. Again, thank you so much for your time and consideration.
Thanks again,
Your Name\"""",  # noqa: E501
]


def construct_meta_prompt(
    instruction: str = None,
    low_quality_example_string: str = None,
    high_quality_example_string: str = None,
    template_type: str = "SIMPLE",
) -> str:
    """Constructs a prompt template for the dataset generator.

    Args:
        instruction: The natural language instruction for the prompt.
        low_quality_example_string: A string representing the low quality examples.
        high_quality_example_string: A string representing the high quality examples.
        template_type: If template_type is COMPLEX, uses the
        COMPLEX_PROMPT_TEMPLATE, if template_type is MIDDLE, uses the
        MIDDLE_PROMPT_TEMPLATE, and if template_type is SIMPLE,
        uses the SIMPLE_PROMPT_TEMPLATE.

    Returns:
        str: A prompt template, where the `instruction` and `examples` fields
            are filled in.
    """
    assert template_type in [
        "SIMPLE",
        "MIDDLE",
        "COMPLEX",
    ], "template_type must be SIMPLE, MIDDLE, or COMPLEX"
    meta_examples = random.sample(META_EXAMPLES, 4)
    example_1, example_2, example_3, example_4 = meta_examples
    if template_type == "COMPLEX":
        return COMPLEX_PROMPT_TEMPLATE.format(
            # The META_PROMPT variable is a shared global variable.
            META_PROMPT=META_PROMPT,
            example_1=example_1,
            example_2=example_2,
            example_3=example_3,
            example_4=example_4,
            instruction=instruction,
            high_quality_example_string=high_quality_example_string,
            low_quality_example_string=low_quality_example_string,
        )
    elif template_type == "MIDDLE":
        return MIDDLE_PROMPT_TEMPLATE.format(
            META_PROMPT=META_PROMPT,
            example_1=example_1,
            example_2=example_2,
            example_3=example_3,
            instruction=instruction,
            high_quality_example_string=high_quality_example_string,
            low_quality_example_string=low_quality_example_string,
        )
    else:
        return SIMPLE_PROMPT_TEMPLATE.format(
            META_PROMPT=META_PROMPT,
            example_1=example_1,
            example_2=example_2,
            instruction=instruction,
            high_quality_example_string=high_quality_example_string,
            low_quality_example_string=low_quality_example_string,
        )



"""Tools for accessing OpenAI's API."""

from __future__ import annotations  # noqa FI58

import asyncio
import json
import logging
import os
import time

import aiolimiter
import openai
import openai.error
import tiktoken
from aiohttp import ClientSession
from tqdm.asyncio import tqdm_asyncio

OPENAI_ERRORS = (
    openai.error.APIError,
    openai.error.Timeout,
    openai.error.RateLimitError,
    openai.error.ServiceUnavailableError,
    openai.error.InvalidRequestError,
    json.decoder.JSONDecodeError,
    AssertionError,
)

ERROR_ERRORS_TO_MESSAGES = {
    openai.error.InvalidRequestError: "OpenAI API Invalid Request: Prompt was filtered",  # noqa E501
    openai.error.RateLimitError: "OpenAI API rate limit exceeded. Sleeping for 10 seconds.",  # noqa E501
    openai.error.APIConnectionError: "OpenAI API Connection Error: Error Communicating with OpenAI",  # noqa E501
    openai.error.Timeout: "OpenAI APITimeout Error: OpenAI Timeout",
    openai.error.ServiceUnavailableError: "OpenAI service unavailable error: {e}",
    openai.error.APIError: "OpenAI API error: {e}",
}


class ChatGPTAgent:
    """A class for accessing OpenAI's ChatCompletion API."""

    def __init__(self, api_key: str | None, model_name: str = "gpt-3.5-turbo"):
        """Initialize ChatGPTAgent with an API key.

        Args:
            api_key: A valid OpenAI API key. Alternatively, set as None and set
                     the environment variable with `export OPENAI_API_KEY=<your key>`.
            model_name: Name fo the OpenAI model to use (by default, gpt-3.5-turbo).
        """
        openai.api_key = api_key if api_key else os.environ["OPENAI_API_KEY"]
        assert openai.api_key is not None and openai.api_key != "", (
            "API key must be provided"
            + " or set the environment variable with `export OPENAI_API_KEY=<your key>`"
        )
        self.model_name = model_name

    def generate_one_openai_chat_completion(
        self,
        prompt: str,
        temperature: float = 0,
        presence_penalty: float = 0,
        frequency_penalty: float = 0,
    ) -> openai.Completion:
        """Generate a chat completion using OpenAI's gpt-3.5-turbo.

        Args:
            prompt: A prompt asking for a response.
            temperature: What sampling temperature to use, between 0 and 2. Higher
                values like 0.8 will make the output more random, while lower values
                like 0.2 will make it more focused and deterministic.
            presence_penalty: Float between -2.0 and 2.0. Positive values penalize new
                tokens based on whether they appear in the text so far, increasing the
                model's likelihood to talk about new topics.
            frequency_penalty: Float between -2.0 and 2.0. Positive values penalize new
                tokens based on their existing frequency in the text so far, decreasing
                the model's likelihood of repeating the same line verbatim.

        Returns:
            A response object.
        """
        response = openai.ChatCompletion.create(
            model=self.model_name,
            messages=[
                {"role": "user", "content": f"{prompt}"},
            ],
            temperature=temperature,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
        )
        return response

    async def generate_batch_openai_chat_completion(
        self,
        prompts: list[str],
        temperature: float = 1,
        responses_per_request: int = 5,
        requests_per_minute: int = 80,
    ) -> list[openai.Completion]:
        """Generate a batch responses from OpenAI Chat Completion API.

        Args:
            prompts: List of prompts to generate from.
            model_config: Model configuration.
            temperature: Temperature to use.
            responses_per_request: Number of responses for each request.
                i.e. the parameter n of OpenAI API call.
            requests_per_minute: Number of requests per minute to allow.

        Returns:
            List of generated responses.
        """
        openai.aiosession.set(ClientSession())
        limiter = aiolimiter.AsyncLimiter(requests_per_minute)

        async def _throttled_openai_chat_completion_acreate(
            model: str,
            messages: list[dict[str, str]],
            temperature: float,
            max_tokens: int,
            n: int,
            top_p: float,
            limiter: aiolimiter.AsyncLimiter,
        ):
            async with limiter:
                for _ in range(3):
                    try:
                        return await openai.ChatCompletion.acreate(
                            model=model,
                            messages=messages,
                            temperature=temperature,
                            max_tokens=max_tokens,
                            n=n,
                            top_p=top_p,
                        )
                    except tuple(ERROR_ERRORS_TO_MESSAGES.keys()) as e:
                        if isinstance(
                            e,
                            (
                                openai.error.ServiceUnavailableError,
                                openai.error.APIError,
                            ),
                        ):
                            logging.warning(
                                ERROR_ERRORS_TO_MESSAGES[type(e)].format(e=e)
                            )
                        elif isinstance(e, openai.error.InvalidRequestError):
                            logging.warning(ERROR_ERRORS_TO_MESSAGES[type(e)])
                            return {
                                "choices": [
                                    {
                                        "message": {
                                            "content": "Invalid Request: Prompt was filtered"  # noqa E501
                                        }
                                    }
                                ]
                            }
                        else:
                            logging.warning(ERROR_ERRORS_TO_MESSAGES[type(e)])
                        await asyncio.sleep(10)
                return {"choices": [{"message": {"content": ""}}]}

        async_responses = [
            _throttled_openai_chat_completion_acreate(
                model="gpt-3.5-turbo",
                messages=[
                    {"role": "user", "content": f"{prompt}"},
                ],
                temperature=temperature,
                max_tokens=500,
                n=responses_per_request,
                top_p=1,
                limiter=limiter,
            )
            for prompt in prompts
        ]
        responses = await tqdm_asyncio.gather(*async_responses)
        # Note: will never be none because it's set, but mypy doesn't know that.
        await openai.aiosession.get().close()
        return responses


def handle_openai_error(e, api_call_counter):
    """Handle OpenAI errors or related errors that the OpenAI API may raise.

    Args:
        e: The error to handle. This could be an OpenAI error or a related
           non-fatal error, such as JSONDecodeError or AssertionError.
        api_call_counter: The number of API calls made so far.

    Returns:
        The api_call_counter (if no error was raised), else raise the error.
    """
    logging.error(e)
    if isinstance(
        e,
        (openai.error.APIError, openai.error.Timeout, openai.error.RateLimitError),
    ):
        # For these errors, OpenAI recommends waiting before retrying.
        time.sleep(1)

    if isinstance(e, OPENAI_ERRORS):
        # For these errors, we can increment a counter and retry the API call.
        return api_call_counter
    else:
        # For all other errors, immediately throw an exception.
        raise e


def count_tokens_from_string(string: str, encoding_name: str = "cl100k_base") -> int:
    """Handle count the tokens in a string with OpenAI's tokenizer.

    Args:
        string: The string to count.
        encoding_name: The name of the tokenizer to use.

    Returns:
        The number of tokens in the string.
    """
    encoding = tiktoken.get_encoding(encoding_name)
    num_tokens = len(encoding.encode(string))
    return num_tokens



"""Import BaseTrainer classes."""
from prompt2model.model_trainer.base import BaseTrainer
from prompt2model.model_trainer.generate import GenerationModelTrainer
from prompt2model.model_trainer.mock import MockTrainer

__all__ = ("MockTrainer", "BaseTrainer", "GenerationModelTrainer")



"""A simple dataset generator that uses OpenAI's GPT-3.5 API."""

from __future__ import annotations  # noqa FI58

import asyncio
import json
import math
import os
import random
from collections import Counter, defaultdict
from dataclasses import dataclass
from pathlib import Path

import openai
from datasets import Dataset
from tqdm import tqdm

from prompt2model.dataset_generator.base import DatasetGenerator, DatasetSplit
from prompt2model.dataset_generator.openai_gpt_template import construct_meta_prompt
from prompt2model.prompt_parser import PromptSpec
from prompt2model.utils import (
    OPENAI_ERRORS,
    ChatGPTAgent,
    count_tokens_from_string,
    get_formatted_logger,
    handle_openai_error,
)

logger = get_formatted_logger("DatasetGenerator")


@dataclass(frozen=True)
class Example:
    """An example from a dataset, containing input and output columns."""

    input_col: str
    output_col: str


class OpenAIDatasetGenerator(DatasetGenerator):
    """A abstract class for NLP dataset generation using OpenAI's GPT-3.5 API."""

    def __init__(
        self,
        api_key: str | None = None,
        max_api_calls: int = None,
        initial_temperature: float = 0.5,
        max_temperature: float = 1.7,
        presence_penalty: float = 0,
        frequency_penalty: float = 0,
        max_batch_size: int = 5,
        responses_per_request: int = 5,
        requests_per_minute: int = 80,
        filter_duplicated_examples: bool = True,
        cache_root: str = "cached_generated_dataset",
    ):
        """Initializes an instance of the OpenAI DatasetGenerator.

        Args:
            api_key: A valid OpenAI API key. If not provided, the environment
                variable OPENAI_API_KEY is used.
            max_api_calls: The maximum number of API calls allowed,
                or None for unlimited.
            initial_temperature: The sampling temperature to use when initializing
                the generation.
            max_temperature: The sampling temperature to use when the generation
                is about to terminate.
            presence_penalty: Value between -2.0 and 2.0 to penalize new tokens
                based on their presence in the text so far. Positive values increase
                the model's likelihood to discuss new topics in generated examples.
            frequency_penalty: Value between -2.0 and 2.0 to penalize new tokens
                based on their frequency in the text. Positive values discourage
                the model from repeating the same line verbatim in generated examples.
            max_batch_size: The maximum number of requests to make in each batch.
            responses_per_request: The number of responses for each request.
            requests_per_minute: The maximum number of requests per minute.
            filter_duplicated_examples: If True, filters duplicated examples,
                using the most-frequent output for each input.
            cache_root: The root directory for caching generated examples.

        Raises:
            AssertionError: If an API key is not provided and set as an environment
            variable, or if the 'max_api_calls' value is not greater than 0.

        Note:
            For the OpenAI GPT-3.5 API, Temperature ranges from 0 to 2. Higher
            values yield more random/diverse outputs with lower quality, while
            lower values produce more deterministic outputs with higher quality.
            We use a strategy to dynamically adjust the temperature from
            initial_temperature to max_temperature during generation.

            We incorporate random few-shot generated examples into the prompt
            to the OpenAI GPT-3.5 API. The initial temperature is set lower to obtain
            high-quality, low-diversity examples. As the number of generated examples
            increases, we gradually have more high-quality examples for in-context
            learning during generation. This allows us to achieve high-quality,
            high-diversity examples later on by using a higher temperature.
        """
        self.api_key: str | None = api_key if api_key else os.environ["OPENAI_API_KEY"]
        assert self.api_key is not None and self.api_key != "", (
            "API key must be provided"
            + " or set the environment variable with `export OPENAI_API_KEY=<your key>`"
        )
        if max_api_calls:
            assert max_api_calls > 0, "max_api_calls must be > 0"
        self.max_api_calls = max_api_calls
        self.api_call_counter = 0
        self.initial_temperature = initial_temperature
        self.max_temperature = max_temperature
        if self.initial_temperature < 0:
            raise ValueError(
                f"initial_temperature must be >= 0, but {self.initial_temperature=}"
            )
        if self.max_temperature > 2.0:
            raise ValueError(
                "max_temperature must be <= 2.0 but {self.max_temperature=}"
            )
        if self.initial_temperature > self.max_temperature:
            raise ValueError(
                f"{self.initial_temperature=} must be <= {self.max_temperature=}"
            )
        self.presence_penalty = presence_penalty
        self.frequency_penalty = frequency_penalty
        self.max_batch_size = max_batch_size
        self.responses_per_request = responses_per_request
        self.requests_per_minute = requests_per_minute
        self.filter_duplicated_examples = filter_duplicated_examples
        self.cache_root = Path(cache_root)

    def construct_prompt(
        self,
        instruction: str,
        few_shot_example_string: str,
        generated_examples: list[Example],
    ) -> str:
        """Generates a prompt string.

        The function generates a prompt string using the provided instruction and
        few_shot_example_string. It also selects random examples from the generated
        dataset to provide additional context for the prompt. At the start of dataset
        generation, it defaults to using the few_shot_example_string.

        The function uses different prompt templates based on the number of selected
        examples from the generated dataset. If the total length of the prompt exceeds
        3500 tokens, repeat the prompt generation process to generate a shorter one.

        Args:
            instruction: The natural language instruction for the prompt.
            few_shot_example_string: A string representing the few-shot examples
                parsed from the user's prompt, which quality is higher than the
                genrated examples.
            generated_examples: A list of currently generated examples.

        Returns:
            The generated prompt string.
        """
        # The random_example_string is a string, which contains several random
        # few-shot examples as demonstrations for the DatasetGenerator. If
        # generated_examples is empty, then the random_example_string
        # is the few-shot examples parsed from the user's prompt.
        while True:
            if len(generated_examples) == 0:
                low_quality_example_string = "N/A\n"
                # Create default low_quality_example_string if generated_examples
                # is empty. few_shot_example_string is the high-quality few-shot
                # examples parsed from the user's prompt. But if user does not
                # provideany examples in the input prompt, the few_shot_example_string
                # will be "N/A"/""/None.
                random_selected_generated_example_num = 0
                # random_selected_generated_example_num is the number of selected
                # random examples from generated_examples that will be used to
                # create the low_quality_example_string. If generated_examples
                # is empty, then random_selected_generated_example_num is 0.
            else:
                # If generated_examples is not empty, low_quality_example_string
                # is sveral random generated examples from generated_examples.

                low_quality_example_string = ""
                random_selected_generated_example_num = random.randint(
                    1, min(len(generated_examples), 10)
                )
                # random_selected_generated_example_num is the number of selected
                # random examples from generated_examples that will
                # be concatenated to create low_quality_example_string.
                random_examples = random.sample(
                    generated_examples, random_selected_generated_example_num
                )
                # If generated_examples is not empty, then choose several
                # random examples from generated_examples to construct
                # new low_quality_example_string.
                for example in random_examples:
                    low_quality_example_string += (
                        f'input="{example.input_col}"\noutput="{example.output_col}"\n'
                    )
            # To increase the diversity of the prompt to DatasetGenerator, create three
            # prompt templates, COMPLEX, MIDDLE, and SIMPLE. The COMPLEX template
            # contains 4 meta examples, the MIDDLE template contains 3 meta examples,
            # and the SIMPLE template contains 2 meta examples.
            template_type_dict = {1: "COMPLEX", 2: "MIDDLE", 0: "SIMPLE"}
            template_type = template_type_dict[
                random_selected_generated_example_num % 3
            ]
            prompt = construct_meta_prompt(
                instruction=instruction,
                low_quality_example_string=low_quality_example_string,
                high_quality_example_string=few_shot_example_string,
                template_type=template_type,
            )
            # The max content length of gpt-3.5-turbo is 4097, so if the
            # generated prompt is longer than 3500, then the prompt
            # should be regenerated.
            if count_tokens_from_string(prompt) < 3500:
                return prompt
            else:
                orginal_input_string = (
                    instruction + few_shot_example_string
                    if few_shot_example_string
                    else instruction
                )
                if count_tokens_from_string(orginal_input_string) > 3500:
                    logger.warning(
                        "The original input prompt is too long. Consider writing a shorter prompt."  # noqa E501
                    )
                continue

    def construct_input_output_map(
        self,
        generated_examples: list[Example],
    ) -> dict[str, Counter]:
        """Constructs a dictionary mapping inputs to `Counter` objects of outputs.

        Args:
            generated_examples: A list of currently generated examples.

        Ideally, each input should have a unique output (one-to-one mapping).
        However, language models may occasionally generate different outputs
        for identical inputs. For instance, given the input “What is the biggest
        city in China?”, it might produce different but correct outputs such as
        “Shanghai” and “The biggest city in China is Shanghai”. At other times,
        it may produce incorrect variations. For the input “What is the Chemical
        symbol of gold?”, the outputs might be “Au”, “Au”, and “AU”, where the
        last one is wrong due to capital letters.

        To address this, OpenAIDataSetGenerator uses a two-step multi-vote
        filtering mechanism. This function represents the first step, creating a
        dictionary to map inputs to a `Counter` of their outputs.

        The function iterates over all the examples, building a dictionary where
        inputs serve as keys and `Counter` objects as values. The `Counter`
        tracks the frequency of each output for a specific input.

        For example:
        input: ["apple", "banana", "apple", "orange", "apple"]
        output: ["A", "B", "A", "O", "D"]

        Then input_output_map value is:
        {
            "apple": Counter({"A": 2, "D": 1}),
            "banana": Counter({"B": 1}),
            "orange": Counter({"O": 1})
        }
        """
        input_output_map: dict[str, Counter] = defaultdict(Counter)

        # Iterate through the examples and construct the mapping.
        for example in generated_examples:
            input_str = example.input_col
            output_str = example.output_col

            # Increment the count of the output for the specific input.
            input_output_map[input_str][output_str] += 1

        # Ensure that the generated_examples list is not empty
        # and the map is constructed correctly.
        if len(generated_examples) != 0:
            assert input_output_map

        return input_output_map

    def apply_multi_vote_to_construct_generated_dataset(
        self, input_output_map: dict[str, Counter]
    ) -> Dataset:
        """Multi-vote to construct generated_dataset from input_output_map.

        Args:
            generated_examples: A list of currently generated examples.

        This method uses multi-vote filtering to create a unique mapping from inputs
        to outputs. The input_col of generated_dataset contains unique inputs,
        while the output_col holds the shortest, most frequent output for the
        corresponding input.

        The function asserts that self.filter_duplicated_examples is True and that
        input_output_map is not None when generated_examples is not
        empty. It then iterates over input_output_map, finding the most frequent
        output for each input. If there are multiple outputs with the highest frequency,
        it selects the shortest one. If there are multiple shortest outputs with the
        highest frequency, it selects the one that comes first in lexicographical
        (alphabetical) order.

        Example:
        Suppose input_output_map is:
        {
            "apple": Counter({"A": 2, "D": 2}),
            "banana": Counter({"B": 2, "C": 1}),
            "orange": Counter({"O": 1})
        }

        The function will produce generated_dataset:
        {
            "input_col": ["apple", "banana", "orange"],
            "output_col": ["A", "B", "O"]
        }

        Note: When generated_examples is empty, both input_output_map
        and generated_dataset will be empty.

        Returns:
            Currently generated dataset with multi-vote filtering applied.
        """
        # Ensure that multi-vote filtering is enabled.
        assert self.filter_duplicated_examples

        filtered_inputs = []
        filtered_outputs = []

        for input_str, output_counter in input_output_map.items():
            # Find the most frequent output count.
            most_common_count = output_counter.most_common(1)[0][1]

            # Get all the outputs that have the most common count.
            most_frequent_outputs = [
                output
                for output, count in output_counter.items()
                if count == most_common_count
            ]

            # Sort the outputs based on their lengths and select
            # the shortest ones. When several outputs have the
            # same length with the highest frequency, they will
            # be sorted in their lexicographical (alphabetical) order.
            most_frequent_outputs.sort(key=len)
            final_output = most_frequent_outputs[0]

            filtered_inputs.append(input_str)
            filtered_outputs.append(final_output)

        # Note that when `generated_examples` is empty,
        # `input_output_map` is None, and `generated_dataset`
        # will also be empty.
        generated_dataset = Dataset.from_dict(
            {"input_col": filtered_inputs, "output_col": filtered_outputs}
        )
        return generated_dataset

    def create_all_examples_dataset_and_generated_dataset(
        self, generated_examples: list[Example]
    ) -> Dataset:
        """Converts generated_examples into generated_dataset.

        Args:
            generated_examples: A list of currently generated examples.

        Depending on the value of self.filter_duplicated_examples, the function either
        constructs a mapping for input-output pairs followed by multi-vote filtering
        to create a Dataset or directly converts the generated examples into a Dataset.

        The function also verifies the presence of data in the input-output map
        and the generated dataset if there are any generated examples and
        self.filter_duplicated_examples is True.

        Lastly, the function stores all generated examples, irrespective of the value
        of self.filter_duplicated_examples, into a Dataset on the disk.

        Returns:
            A dataset of all the generated exmaples and the currently generated
            dataset. If filter_duplicated_examples is True, multi-vote filtering is
            performed. Else, the generated examples are directly converted into
            a Dataset.
        """
        # Convert all generated examples into a Dataset.
        all_generated_examples_dataset = Dataset.from_dict(
            {
                "input_col": [example.input_col for example in generated_examples],
                "output_col": [example.output_col for example in generated_examples],
            }
        )

        if self.filter_duplicated_examples:
            # When filtering duplicated examples is
            # enabled, perform multi-vote filtering.
            input_output_map = self.construct_input_output_map(generated_examples)
            generated_dataset = self.apply_multi_vote_to_construct_generated_dataset(
                input_output_map
            )

            # Ensure that the input-output map and
            # the generated dataset are not empty.
            if len(generated_examples) != 0:
                assert input_output_map is not None
                assert generated_dataset is not None
        else:
            # When filtering duplicated examples is not enabled,
            # use all_generated_examples_dataset directly.
            generated_dataset = all_generated_examples_dataset

        # If there are generated examples, the
        # generated_dataset should not be empty.
        if len(generated_examples) != 0:
            assert len(generated_dataset) > 0

        return all_generated_examples_dataset, generated_dataset

    def compute_batch_size(
        self, expected_num_examples: int, generated_dataset: Dataset
    ) -> int:
        """Computes the batch size for OpenAI API calls in a batch.

        The batch size is determined based on the remaining number of examples to be
        generated and the number of responses per request. The function also respects
        the maximum limit of API calls if it is set.

        Args:
            expected_num_examples: The total number of examples expected to be
            generated for the current dataset split. Note that if max_api_calls is not
            set, the actual number of generated examples can be slightly higher due
            to each API call returning `responses_per_request` examples.
            generated_dataset: Currently generated dataset.

        Returns:
            The batch size for the next batch of OpenAI API calls with zeno-build.
        """
        # If max_api_calls is not set, make it equivalent to the batch size
        max_api_calls = (
            self.max_batch_size
            if self.max_api_calls is None
            else self.max_api_calls - self.api_call_counter
        )
        batch_size = min(
            self.max_batch_size,
            math.ceil(
                (
                    (expected_num_examples - len(generated_dataset))
                    / self.responses_per_request
                )
            ),
            max_api_calls,
        )
        # Ensure that the batch_size is a positive value.
        assert batch_size > 0
        return batch_size

    def extract_responses(
        self, completions: list[openai.Completion], generated_examples: list[Example]
    ) -> list[Example]:
        """Extracts the generated sample and annotation from an OpenAI API response.

        Args:
            completions: A list of Completion objects returned by the OpenAI API.
            Each API call returns a number of completion objects equivalent to
            `responses_per_request`. The default `responses_per_request` = 5.
            generated_examples: Currently generated examples of DatasetGenerator.

        This function iterates through the provided completions, attempting to
        extract and parse the content of each completion as a JSON object. It
        then checks for the presence of `input` and `output` keys in the JSON
        object. If either is missing, the completion is discarded.

        For valid completions, the function instantiates a class
        with `input_col` and `output_col` fields, representing the generated
        example and label strings respectively. The `example` is then added
        to generated_examples.

        Note: The function process `batch_size * responses_per_request`
        responses at a time.

        Example:
            Given a list of two completion objects: [completion_1, completion_2],
            where:
            completion_1.choices = [
                {"message": {"content": '{"input": "1", "output": "a"}'}},
                {"message": {"content": '{"input": "1", "output": "b"}'}},
                {"message": {"content": '{"input": "1", "output": "a"}'}},
            ]
            completion_2.choices = [
                {"message": {"content": '{"input": "1", "output": "c"}'}},
                {"message": {"content": '{"input": "2", "output": "a"}'}},
                {"message": {"content": '{"input": "2", "output": "b"}'}},
            ]

            The function will create 'example' namedtuples:
            Example(input_col="1", output_col="a")
            Example(input_col="1", output_col="b")
            Example(input_col="1", output_col="a")
            Example(input_col="1", output_col="c")
            Example(input_col="2", output_col="a")
            Example(input_col="2", output_col="b")

            It will then append them to generated_examples.

        Returns:
            A list of `Example` objects.
                Each API call will return `responses_per_request` completion objects.
                If the response is a valid JSON object, create a namedtuple called
                `example` and append it to generated_examples. `example` consists
                of `input_col` and`output_col`, where:
                - input_col is the generated example string extracted from the response.
                - output_col is the generated label string extracted from the response.
                If the response is not a valid JSON object, discard it.
                There is responses_per_request * len(completions) responses at a time.
        """
        for completion in completions:
            try:
                for choice in completion.choices:
                    try:
                        response_json = json.loads(choice["message"]["content"])
                    except Exception:
                        logger.warning(f"Error happened parsing API choice: {choice}")
                        continue
                        # If the response is not a valid JSON object, discard it.
                    required_keys = ["input", "output"]
                    missing_keys = [
                        key for key in required_keys if key not in response_json
                    ]
                    if len(missing_keys) != 0:
                        logger.warning(
                            f'API response must contain {", ".join(required_keys)} keys'
                        )
                        continue
                    input = str(response_json["input"]).strip()
                    output = str(response_json["output"]).strip()
                    if input != "" and output != "":
                        generated_examples.append(Example(input, output))
                    else:
                        logger.info(
                            "Empty input or output ditected. Discard this example."
                        )
                        continue
                    logger.info(f"input: \n\n{input}\n\n")
                    logger.info(f"output: \n\n{output}\n\n")
            except Exception:
                logger.warning(
                    f"Error happened when parsing API completion: {completion}"
                )
                continue
        return generated_examples

    async def generate_responses(
        self,
        chat_api: ChatGPTAgent,
        generated_dataset: Dataset,
        expected_num_examples: int,
        prompts: list[str],
    ) -> list[openai.Completion]:
        """Asynchronously generates responses using the GPT-3.5 API.

        Args:
            chat_api: ChatGPTAgent to generate responses.
            generated_dataset: Currently generated dataset.
            expected_num_examples: The number of examples expected
                to be generated.
            prompts: A list of prompts to generate responses.

        The temperature for generating responses dynamically adjusts
        based on the size of the generated dataset. As the dataset
        grows, the dynamic temperature gradually increases,
        approaching the max_temperature.

        To prevent potential round-off errors in Python, the dynamic
        temperature is rounded within the range [0, 2.0].

        Returns:
            A list of openai.Completion.
        """
        # Calculate the dynamic temperature based
        # on the size of the generated dataset
        dynamic_temperature = (self.max_temperature - self.initial_temperature) * len(
            generated_dataset
        ) / expected_num_examples + self.initial_temperature

        # Ensure the dynamic temperature is within the range [0, 2.0]
        clipped_temperature = max(0.0, min(2.0, dynamic_temperature))
        responses = await chat_api.generate_batch_openai_chat_completion(
            prompts,
            temperature=clipped_temperature,
            responses_per_request=self.responses_per_request,
            requests_per_minute=self.requests_per_minute,
        )
        return responses

    def generate_dataset_split(
        self,
        prompt_spec: PromptSpec,
        expected_num_examples: int,
        split: DatasetSplit,
    ) -> Dataset:
        """Generates a dataset split using GPT-3.5.

        This method iteratively makes API calls to GPT-3.5 to generate a dataset split.
        Each API call yields a batch of responses. From these responses, new examples
        are extracted and added to 'generated_examples'. The process continues
        until the desired number of examples is reached, or the maximum limit on API
        calls is reached. If an error occurs during an API call, the error is handled
        appropriately, and the API call counter is adjusted.

        Args:
            prompt_spec: PromptParser to be used for generating examples.
            expected_num_examples: The number of examples expected to be
                generated. If the maximum limit on API calls is not set, the actual
                number of generated examples can be slightly higher due to each
                API call returning `responses_per_request` examples.
            split: The dataset split (e.g., train, validation, test) for which the
                examples are being generated.

        Returns:
            The generated dataset split.
        """
        # Refresh the relevant data structures for the new split.
        self.cache_root.mkdir(parents=True, exist_ok=True)
        examples_cache_path = Path(
            self.cache_root / f"generated_examples_{split.value}"
        )
        dataset_cache_path = Path(self.cache_root / f"generated_dataset_{split.value}")

        if examples_cache_path.exists():
            # If cache exists, load generated examples from disk.
            logger.info(f"Loading cache from {str(examples_cache_path)}.")
            all_generated_examples_dataset = Dataset.load_from_disk(examples_cache_path)
            generated_examples = [
                Example(input_col=ex["input_col"], output_col=ex["output_col"])
                for ex in all_generated_examples_dataset
            ]
        else:
            # Initialize data structures for a new split.
            generated_examples = []

        pbar = tqdm(total=expected_num_examples, desc="Generating examples")
        chat_api = ChatGPTAgent(self.api_key)

        while True:
            # Each API call will return `responses_per_request` completion
            # objects. The upper bound of the length of the generated dataset
            # is expected_num_examples + responses_per_request.
            try:
                # Convert the generated examples into a
                # Dataset and update the progress bar.
                (
                    all_generated_examples_dataset,
                    generated_dataset,
                ) = self.create_all_examples_dataset_and_generated_dataset(
                    generated_examples
                )
                all_generated_examples_dataset.save_to_disk(examples_cache_path)
                generated_dataset.save_to_disk(dataset_cache_path)
                pbar.update(len(generated_dataset))

                if self.max_api_calls and self.api_call_counter >= self.max_api_calls:
                    logger.warning("Maximum number of API calls reached.")
                    break
                elif len(generated_dataset) >= expected_num_examples:
                    break
                else:
                    # Compute the batch size for the next API call.
                    batch_size = self.compute_batch_size(
                        expected_num_examples, generated_dataset
                    )
                    self.api_call_counter += batch_size

                    # Generate prompts for the batch call.
                    prompts = [
                        self.construct_prompt(
                            instruction=prompt_spec.instruction,
                            few_shot_example_string=prompt_spec.examples,
                            generated_examples=generated_examples
                            if not self.filter_duplicated_examples
                            else [
                                Example(each["input_col"], each["output_col"])
                                for each in generated_dataset
                            ],
                        )
                        for _ in range(batch_size)
                    ]

                    loop = asyncio.get_event_loop()
                    responses = loop.run_until_complete(
                        self.generate_responses(
                            chat_api=chat_api,
                            generated_dataset=generated_dataset,
                            expected_num_examples=expected_num_examples,
                            prompts=prompts,
                        )
                    )

                    # Extract the responses and add new examples to the dataset.
                    generated_examples = self.extract_responses(
                        responses, generated_examples
                    )
            except OPENAI_ERRORS as e:
                # Handle OpenAI API errors and adjust the API call counter.
                self.api_call_counter = handle_openai_error(e, self.api_call_counter)
        return generated_dataset



"""A base class for dataset processor."""

from abc import ABC, abstractmethod
from functools import partial

import datasets


class BaseProcessor(ABC):
    """A base class for post-processing datasets."""

    def __init__(self, has_encoder: bool, eos_token: str) -> None:
        """Initialize the `BaseProcessor`.

        Args:
            has_encoder: Whether the retrieved model has an encoder.
                Encoder-decoder model like T5 has two model inputs.
                Decoder-only model like GPT only has one model input, thus
                `model_input` should be added with the `output_col`.
            eos_token: The end-of-sentence token of the tokenizer.
        """
        self.has_encoder = has_encoder
        self.eos_token = eos_token

    @staticmethod
    @abstractmethod
    def post_process_example(
        example: dict,
        instruction: str,
        task_id: int,
        has_encoder: bool,
        dataset_split: str,
        eos_token: str,
    ) -> dict:
        """Modifies the input column of a given example dictionary.

        Args:
            example: A dictionary representing an example.
            instruction: The instruction used as a prefix to explain the task.
            task_id: A tag marking which dataset (from dataset_dicts) this
                example comes from. Used for multi-task training.
            has_encoder: Whether the retrieved model has an encoder.
            dataset_split: The split of the example, i.e. train/val/test.
            eos_token: The end-of-sentence token of the tokenizer.

        Returns:
            A dictionary with `model_input` as the input to models
            and `model_output` as the expected output of models.
        """

    def process_dataset_dict(
        self, instruction: str, dataset_dicts: list[datasets.DatasetDict]
    ) -> list[datasets.DatasetDict]:
        """Post-process a list of DatasetDicts.

        Args:
            instruction: The instruction used as a prefix to explain the task.
            dataset_dicts: A list of DatasetDicts (generated or retrieved).

        Returns:
            A list of DatasetDicts, all examples are converted into text2text fashion.
            Note that if any example contain empty `input_col` or `output_col`,
            it will be discarded.
        """
        modified_dataset_dicts = []

        def filter_empty_strings(example: dict) -> bool:
            """Filter to exclude examples with empty 'input_col' or 'output_col'.

            Args:
                example: A dictionary representing a single example in the dataset.

            Returns:
                bool: True if both 'input_col' and 'output_col' are non-empty strings,
                    False otherwise.

            Raises:
                AssertionError: If no 'input_col' or 'output_col' inside the example.
            """
            assert (
                "input_col" in example and "output_col" in example
            ), "Example dictionary must have 'input_col' and 'output_col' keys."
            # Check if 'input_col' and 'output_col' are both non-empty strings
            return bool(example["input_col"]) and bool(example["output_col"])

        for task_id, dataset_dict in enumerate(dataset_dicts):
            modified_dataset_dict = {}
            for dataset_split in list(dataset_dict.keys()):
                mapping_function = partial(
                    self.post_process_example,
                    instruction=instruction,
                    task_id=task_id,
                    has_encoder=self.has_encoder,
                    dataset_split=dataset_split,
                    eos_token=self.eos_token,
                )
                modified_dataset_dict[dataset_split] = (
                    dataset_dict[dataset_split]
                    .filter(filter_empty_strings)
                    .map(mapping_function)
                )
            modified_dataset_dict = datasets.DatasetDict(modified_dataset_dict)
            modified_dataset_dicts.append(modified_dataset_dict)
        return modified_dataset_dicts



"""An interface for model selection."""

from prompt2model.model_retriever import ModelRetriever
from prompt2model.prompt_parser import PromptSpec


class MockModelRetriever(ModelRetriever):
    """Select a fixed model from among a set of hyperparameter choices."""

    def __init__(self, fixed_model_name: str):
        """Initialize a dummy retriever that returns a fixed model name."""
        self.fixed_model_name = fixed_model_name

    def retrieve(
        self,
        prompt: PromptSpec,
    ) -> str:
        """Select an arbitrary, fixed model from HuggingFace.

        Args:
            prompt: A prompt to use to select relevant models.

        Return:
            A relevant model's HuggingFace name.
        """
        return self.fixed_model_name



"""A trainer class to train generation models."""

from __future__ import annotations  # noqa FI58

import os
from itertools import takewhile
from typing import Any

import datasets
import torch
import torch.nn as nn
import transformers
from datasets import concatenate_datasets
from transformers import Seq2SeqTrainer, Seq2SeqTrainingArguments

from prompt2model.model_trainer.base import BaseTrainer
from prompt2model.model_trainer.callback import ValidationCallback
from prompt2model.utils import get_formatted_logger, seed_generator

logger = get_formatted_logger("ModelTrainer")

os.environ["TOKENIZERS_PARALLELISM"] = "false"


class GenerationModelTrainer(BaseTrainer):
    """Trainer for T5 type (encoder-decoder) model and GPT type (deocder-only) model."""

    def __init__(
        self,
        pretrained_model_name: str,
        has_encoder: bool,
        executor_batch_size: int = 10,
        tokenizer_max_length: int = 512,
        sequence_max_length: int = 1024,
    ):
        """Initializes a new instance of GenerationModelTrainer.

        Args:
            pretrained_model_name: HuggingFace pre-trained model name.
                Only supported encoder-decoder model or atuoregressive model.
            has_encoder: Whether the model has an encoder.
                If True, it's a T5-type model (encoder-decoder transformer).
                If fasle, it's a GPT-type model (atuoregressive transformer).
            executor_batch_size: The batch size for model executor to
                make predictions.
            tokenizer_max_length: The maximum sentence length the tokenizer
                is allowed to generate.
            sequence_max_length: The maximum number of tokens the model is
                allowed to generate when being evaluated on validation dataset.
                Note that sequence_max_length might be scaled in the ModelExecutor
                if it exceeds the model's max_embedding.
        """
        self.has_encoder = has_encoder
        self.tokenizer_max_length = tokenizer_max_length
        self.sequence_max_length = sequence_max_length
        self.executor_batch_size = executor_batch_size
        if self.tokenizer_max_length is None:
            logger.warning(
                (
                    "Set the tokenizer_max_length is preferable for finetuning model,"
                    " which saves the cost of training."
                )
            )
        if self.has_encoder:
            self.model = transformers.AutoModelForSeq2SeqLM.from_pretrained(
                pretrained_model_name
            )
            self.tokenizer = transformers.AutoTokenizer.from_pretrained(
                pretrained_model_name
            )
        else:
            self.model = transformers.AutoModelForCausalLM.from_pretrained(
                pretrained_model_name
            )
            self.tokenizer = transformers.AutoTokenizer.from_pretrained(
                pretrained_model_name, padding_side="left"
            )

        if self.tokenizer.pad_token is None:
            self.tokenizer.pad_token = self.tokenizer.eos_token
        if self.model.config.pad_token_id is None:
            self.model.config.pad_token_id = self.tokenizer.eos_token_id

        # self.validation_callback is used for evaluate the model on
        # the validation dataset after each epoch.
        self.validation_callback = None
        self.training_seed = seed_generator.get_seed()

    def get_left_padding_length(cls, input_list, padding_token_id):
        """Get the left prefix length of the input list.

        Args:
            input_list: A list with the format of [prefix, ..., prefix, Others].
                The GPT tokenizer uses left padding.
            padding_token_id: The prefix of the input list.

        Returns:
            The length of [prefix, ..., prefix] in [prefix, ..., prefix, Others]
        """
        return len(list(takewhile(lambda x: x == padding_token_id, input_list)))

    def get_right_padding_length(cls, input_ids, padding_token_id):
        """Get the left prefix length of the input list.

        Args:
            input_list: A list with the format of [Others, suffix, ..., suffix].
                The T5 tokenizer uses right padding.
            padding_token_id: The suffix of the input list.

        Returns:
            The length of [suffix, ..., suffix] in [Others, suffix, ..., suffix].
        """
        # Reverse the input_ids to get the length of right padding.
        suffix_length = cls.get_left_padding_length(input_ids[::-1], padding_token_id)
        return suffix_length

    def tokenize_dataset(
        self, dataset: datasets.Dataset, shuffle: bool = True
    ) -> datasets.Dataset:
        """Tokenize the training/validation dataset.

        Args:
            dataset: Dataset.dataset with model_input and model_output columns.
            shuffle: Whether to shuffle the dataset.

        Returns:
            A `datasets.Dataset` object containing the preprocessed data with:
                "input_ids": numerical representations of input sequences to the model.
                "attention_mask": Mask to avoid performing attention on padding token
                    indices. Mask values selected in [0, 1]: 1 for tokens that are not
                    masked, 0 for masked tokens.
                "labels": Labels for language modeling. Indices are selected in
                    [-100, 0, ..., config.vocab_size - 1]. All labels set to -100 are
                    ignored (masked), the loss is only computed for labels in
                    [0, ..., config.vocab_size - 1].

                Note that -100 is the default ignore index for labels when computing
                    loss. You can check it by:
                    from torch import nn
                    loss_function = nn.CrossEntropyLoss()
                    IGNORE_INDEX = loss_function.ignore_index
        """
        if shuffle:
            dataset = dataset.shuffle(seed=seed_generator.get_seed())
        inputs = dataset["model_input"]
        outputs = dataset["model_output"]
        longest_input = max(inputs, key=len)
        longest_output = max(outputs, key=len)
        if self.tokenizer_max_length is not None and (
            len(self.tokenizer.tokenize(longest_input)) > self.tokenizer_max_length
            or len(self.tokenizer.tokenize(longest_output)) > self.tokenizer_max_length
        ):
            logger.warning(
                (
                    "Truncation happened when tokenizing dataset."
                    " Consider increasing the tokenizer_max_length if possible."
                    " Otherwise, truncation may lead to unexpected results."
                )
            )
        input_encodings = self.tokenizer.batch_encode_plus(
            inputs,
            truncation=True,
            max_length=self.tokenizer_max_length,
            padding=True,
        )
        output_encodings = self.tokenizer.batch_encode_plus(
            outputs,
            truncation=True,
            max_length=self.tokenizer_max_length,
            padding=True,
        )

        labels = []
        loss_function = nn.CrossEntropyLoss()
        IGNORE_INDEX = loss_function.ignore_index
        if not self.has_encoder:
            length_of_input_encoding_ids_with_padding = len(
                input_encodings["input_ids"][0]
            )
            length_of_output_encoding_ids_with_padding = len(
                output_encodings["input_ids"][0]
            )
            for idx, input_id in enumerate(input_encodings["input_ids"]):
                output_encoding_id = output_encodings["input_ids"][idx]
                length_of_padding_in_output_encoding_id = self.get_left_padding_length(
                    output_encoding_id, self.model.config.pad_token_id
                )
                # The IGNORE_INDEX is ignored for loss compute in Autoregressive model.
                # Reference: https://huggingface.co/docs/transformers/model_doc/gpt2#transformers.GPT2DoubleHeadsModel.forward.labels # noqa E501
                length_of_output_encoding_id_without_padding = (
                    length_of_output_encoding_ids_with_padding
                    - length_of_padding_in_output_encoding_id
                )
                assert (
                    length_of_output_encoding_id_without_padding != 0
                ), "One of the model_output is empty."
                label = [IGNORE_INDEX] * (
                    length_of_input_encoding_ids_with_padding
                    - length_of_output_encoding_id_without_padding
                ) + input_id[-length_of_output_encoding_id_without_padding:]
                assert (
                    len(label)
                    == length_of_input_encoding_ids_with_padding
                    == len(input_id)
                )
                labels.append(label)
        else:
            # For T5 model, right padding token id should ignored by the loss
            # function. In PyTorch and Tensorflow, this can be done by replacing
            # them with IGNORE_INDEX, which is the ignore_index of the
            # CrossEntropyLoss as demonstrated before.
            # Reference: https://huggingface.co/docs/transformers/v4.30.0/en/model_doc/t5#training # noqa E501
            for output_encoding_id in output_encodings["input_ids"]:
                length_of_right_padding_in_output_encoding_id = (
                    self.get_right_padding_length(
                        output_encoding_id, self.tokenizer.pad_token_id
                    )
                )
                label = (
                    (
                        output_encoding_id[
                            :-length_of_right_padding_in_output_encoding_id
                        ]
                        + [IGNORE_INDEX] * length_of_right_padding_in_output_encoding_id
                    )
                    if length_of_right_padding_in_output_encoding_id != 0
                    else output_encoding_id
                )
                assert len(label) == len(output_encoding_id)
                labels.append(label)

        preprocessed_dict = {
            "input_ids": input_encodings["input_ids"],
            "attention_mask": input_encodings["attention_mask"],
            "labels": labels,
        }
        return datasets.Dataset.from_dict(preprocessed_dict)

    def train_model(
        self,
        hyperparameter_choices: dict[str, Any],
        training_datasets: list[datasets.Dataset],
        validation_datasets: list[datasets.Dataset] | None = None,
    ) -> tuple[transformers.PreTrainedModel, transformers.PreTrainedTokenizer]:
        """Train a text generation model.

        Args:
            hyperparameter_choices: A dictionary of hyperparameters for training.
            training_datasets: Training datasets with `input_col` and `model_output`.
            validation_datasets: Validation datasets during training. If not provided,
                15% of training data will be spilt from training_datasets to validate.

        Returns:
            A trained HuggingFace model and tokenizer.
        """
        hyperparameter_choices_keys = set(hyperparameter_choices.keys())
        supported_keys = {
            "output_dir",
            "logging_steps",
            "evaluation_strategy",
            "save_strategy",
            "num_train_epochs",
            "per_device_train_batch_size",
            "warmup_steps",
            "weight_decay",
            "logging_dir",
            "learning_rate",
            "test_size",
        }
        assert hyperparameter_choices_keys.issubset(
            supported_keys
        ), f"Only support {supported_keys} as training parameters."
        training_args = Seq2SeqTrainingArguments(
            output_dir=hyperparameter_choices.get("output_dir", "./result"),
            logging_steps=hyperparameter_choices.get("logging_steps", 1),
            save_strategy=hyperparameter_choices.get("save_strategy", "no"),
            num_train_epochs=hyperparameter_choices.get("num_train_epochs", 10),
            per_device_train_batch_size=hyperparameter_choices.get(
                "per_device_train_batch_size", 100
            ),
            warmup_steps=hyperparameter_choices.get("warmup_steps", 0),
            weight_decay=hyperparameter_choices.get("weight_decay", 0.01),
            logging_dir=hyperparameter_choices.get("logging_dir", "./result"),
            learning_rate=hyperparameter_choices.get("learning_rate", 1e-4),
            predict_with_generate=True,
        )
        evaluation_strategy = hyperparameter_choices.get("evaluation_strategy", "epoch")
        if evaluation_strategy == "epoch":
            evaluate_after_epoch = True
        elif evaluation_strategy == "no":
            logger.info(
                "The trainer doesn't set the evaluation strategy, the evaluation will be skipped."  # noqa E501
            )
            evaluate_after_epoch = False
        else:
            logger.warning(
                (
                    "Only `epoch` evaluation strategy is supported"
                    + ", the evaluation strategy will be set to evaluate_after_epoch."
                )
            )
            evaluate_after_epoch = True

        concatenated_training_dataset = concatenate_datasets(training_datasets)

        if evaluate_after_epoch is True:
            if validation_datasets is None:
                if not self.has_encoder:
                    # The validation dataset for autoregressive model is missing.
                    logger.warning(
                        (
                            (
                                "The validation split for autoregressive model is missing"  # noqa E501
                                + ", which should not contain labels as the training spilt."  # noqa E501
                                + " Thus this evaluation will be skipped."
                            )
                        )
                    )
                    train_dataset = self.tokenize_dataset(concatenated_training_dataset)
                    val_dataset = None
                    evaluate_after_epoch = False
                else:
                    # The validation dataset for encoder-decoder model is missing.
                    logger.warning(
                        (
                            "The validation split for encoder-decoder model is missing."  # noqa E501
                            + " The training dataset will be split to create the validation dataset."  # noqa E501
                        )
                    )
                    test_size = hyperparameter_choices.get("test_size", 0.15)
                    assert (
                        len(concatenated_training_dataset) > 1
                    ), "Dataset should be larger than 1 to make train/test split."
                    splitted_dataset = concatenated_training_dataset.train_test_split(
                        test_size=test_size, seed=self.training_seed
                    )
                    train_dataset = self.tokenize_dataset(splitted_dataset["train"])
                    # The training dataset will be tokenized to train the model.
                    # We evaluate the model on the validation dataset in the
                    # callback with the model executor and model evaluator,
                    # so the validation dataset should not be pre-tokenized here.
                    val_dataset = splitted_dataset["test"]
            else:
                # the training dataset will be tokenized to train the model.
                # But we evaluate the model on the validation dataset in the
                # call back with the model executor and model evaluator,
                # the validation dataset should not be tokenized.
                train_dataset = self.tokenize_dataset(concatenated_training_dataset)
                val_dataset = concatenate_datasets(validation_datasets)
        else:
            if validation_datasets:
                logger.warning(
                    "The validation dataset is provided, but the evaluation is skipped."  # noqa E501
                )
            train_dataset = self.tokenize_dataset(concatenated_training_dataset)
        trainer = Seq2SeqTrainer(
            model=self.model,
            args=training_args,
            train_dataset=train_dataset,
            data_collator=transformers.DataCollatorForSeq2Seq(tokenizer=self.tokenizer)
            if self.has_encoder
            else transformers.DataCollatorForLanguageModeling(
                tokenizer=self.tokenizer, mlm=False
            ),
            optimizers=[
                torch.optim.AdamW(
                    params=self.model.parameters(), lr=training_args.learning_rate
                ),
                None,
            ],
        )

        if evaluate_after_epoch:
            assert val_dataset is not None, "Validation dataset is None"
            self.validation_callback = ValidationCallback(
                trainer,
                self.tokenizer,
                val_dataset,
                executor_batch_size=self.executor_batch_size,
                tokenizer_max_length=self.tokenizer_max_length,
                sequence_max_length=self.sequence_max_length,
            )
            trainer.add_callback(self.validation_callback)

        # Train the model
        trainer.train()
        # Return the trained model and tokenizer
        return self.model, self.tokenizer



"""Import evaluator classes."""
from prompt2model.model_evaluator.base import ModelEvaluator
from prompt2model.model_evaluator.mock import MockEvaluator
from prompt2model.model_evaluator.seq2seq import Seq2SeqEvaluator

__all__ = ("MockEvaluator", "ModelEvaluator", "Seq2SeqEvaluator")



"""Tools for encoding and serializing a search index with a contextual encoder."""

from __future__ import annotations  # noqa FI58

import json
import os
import pickle
import tempfile
from contextlib import nullcontext

import numpy as np
import torch
from tevatron.arguments import DataArguments
from tevatron.data import EncodeCollator, EncodeDataset
from tevatron.datasets import HFCorpusDataset, HFQueryDataset
from tevatron.modeling import DenseModelForInference
from torch.utils.data import DataLoader
from transformers import AutoConfig, AutoTokenizer, PreTrainedTokenizerBase


def load_tevatron_model(
    model_name_or_path: str, model_cache_dir: str | None = None
) -> tuple[DenseModelForInference, PreTrainedTokenizerBase]:
    """Load a Tevatron model from a model name/path.

    Args:
        model_name_or_path: The HuggingFace model name or path to the model.
        model_cache_dir: The directory to cache the model.

    Returns:
        A Tevatron dense retrieval model and its associated tokenizer.
    """
    config = AutoConfig.from_pretrained(
        model_name_or_path,
        cache_dir=model_cache_dir,
    )
    tokenizer = AutoTokenizer.from_pretrained(
        model_name_or_path,
        cache_dir=model_cache_dir,
        use_fast=False,
    )
    model = DenseModelForInference.build(
        model_name_or_path=model_name_or_path,
        config=config,
        cache_dir=model_cache_dir,
    )
    return model, tokenizer


def encode_text(
    model_name_or_path: str,
    file_to_encode: str | None = None,
    text_to_encode: list[str] | str | None = None,
    encode_query: bool = False,
    encoding_file: str | None = None,
    max_len: int = 400,
    device: torch.device = torch.device("cpu"),
    dataloader_num_workers: int = 0,
    model_cache_dir: str | None = None,
    data_cache_dir: str = "~/.cache/huggingface/datasets",
    batch_size=8,
    fp16: bool = False,
) -> np.ndarray:
    """Encode a query or documents.

    This code is mostly duplicated from tevatron/driver/encode.py in the Tevatron
    repository.

    Args:
        model_name_or_path: The HuggingFace model name or path to the model.
        file_to_encode: JSON or JSONL file containing `"text"` fields to encode.
        text_to_encode: String or list of strings to encode.
        encode_query: Whether or not we are encoding a query or documents.
        encoding_file: If given, store the encoded data in this file.
        max_len: Truncate the input to this length (in tokens).
        device: Device that Torch will use to encode the text.
        dataloader_num_workers: Number of workers to use for the dataloader.
        model_cache_dir: The directory to cache the model.
        data_cache_dir: The directory to cache the tokenized dataset.
        batch_size: Batch size to use for encoding.
        fp16: Whether or not to run inference in fp16 for more-efficient encoding.

    Returns:
        A numpy array of shape `(expected_num_examples, embedding_dim)` containing text
        encoded by the specified model.
    """
    model, tokenizer = load_tevatron_model(model_name_or_path, model_cache_dir)

    if file_to_encode is None and text_to_encode is None:
        raise ValueError("Must provide either a dataset file or text to encode.")
    elif file_to_encode is not None and text_to_encode is not None:
        raise ValueError("Provide either a dataset file or text to encode, not both.")

    with tempfile.TemporaryDirectory() as temp_dir:
        if text_to_encode is not None:
            if isinstance(text_to_encode, str):
                text_to_encode = [text_to_encode]
            with open(
                os.path.join(temp_dir, "text_to_encode.json"), "w"
            ) as temporary_file:
                text_rows = [
                    {"text_id": i, "text": text}
                    for i, text in enumerate(text_to_encode)
                ]
                json.dump(text_rows, temporary_file)
                file_to_encode = temporary_file.name
                temporary_file.close()

        data_args = DataArguments(
            encoded_save_path=encoding_file,
            encode_in_path=file_to_encode,
            encode_is_qry=encode_query,
            data_cache_dir=data_cache_dir,
        )
        if encode_query:
            data_args.q_max_len = max_len
            hf_dataset = HFQueryDataset(
                tokenizer=tokenizer,
                data_args=data_args,
                cache_dir=data_args.data_cache_dir or model_cache_dir,
            )
        else:
            data_args.p_max_len = max_len
            hf_dataset = HFCorpusDataset(
                tokenizer=tokenizer,
                data_args=data_args,
                cache_dir=data_args.data_cache_dir or model_cache_dir,
            )

        encode_dataset = EncodeDataset(
            hf_dataset.process(1, 0), tokenizer, max_len=max_len
        )

        encode_loader = DataLoader(
            encode_dataset,
            batch_size=batch_size,
            collate_fn=EncodeCollator(
                tokenizer, max_length=max_len, padding="max_length"
            ),
            shuffle=False,
            drop_last=False,
            num_workers=dataloader_num_workers,
        )
        encoded = []
        lookup_indices = []
        model = model.to(device)
        model.eval()

        for batch_ids, batch in encode_loader:
            lookup_indices.extend(batch_ids)
            with torch.cuda.amp.autocast() if fp16 else nullcontext():
                with torch.no_grad():
                    for k, v in batch.items():
                        batch[k] = v.to(device)
                    if data_args.encode_is_qry:
                        model_output = model(query=batch)
                        encoded.append(model_output.q_reps.cpu().detach().numpy())
                    else:
                        model_output = model(passage=batch)
                        encoded.append(model_output.p_reps.cpu().detach().numpy())

        encoded = np.concatenate(encoded)

        if encoding_file:
            with open(encoding_file, "wb") as f:
                pickle.dump((encoded, lookup_indices), f)

        return encoded



"""Script to run the dataset retriever in isolation."""
from prompt2model.dataset_retriever import DescriptionDatasetRetriever
from prompt2model.prompt_parser import MockPromptSpec, TaskType

if __name__ == "__main__":
    prompt_spec = MockPromptSpec(TaskType.TEXT_GENERATION)
    prompt = """Your task is to generate an answer to a natural question. In this task, the input is a string that consists of both a question and a context passage. The context is a descriptive passage related to the question and contains the answer. And the question can range from Math, Cultural, Social, Geometry, Biology, History, Sports, Technology, Science, and so on."""  # noqa E501
    prompt_spec._instruction = prompt

    retriever = DescriptionDatasetRetriever()
    retriever.retrieve_dataset_dict(
        prompt_spec, blocklist=["squad", "stanford question answering"]
    )



"""An interface for automatically evaluate Seq2Seq generation model."""

from __future__ import annotations  # noqa FI58

from typing import Any

import datasets
import evaluate
import numpy as np

from prompt2model.model_evaluator.base import ModelEvaluator
from prompt2model.model_executor import ModelOutput
from prompt2model.utils import get_formatted_logger

logger = get_formatted_logger("ModelEvaluator")


class Seq2SeqEvaluator(ModelEvaluator):
    """An evaluator computing `chr_f++`, `Exact Match` and `Embedding Distance`."""

    def evaluate_model(
        self,
        dataset: datasets.Dataset,
        gt_column: str,
        predictions: list[ModelOutput],
        model_input_column: str | None = None,
        metrics: list[evaluate.Metric] | None = None,
        encoder_model_name: str = "xlm-roberta-base",
    ) -> dict[str, Any]:
        """Evaluate a model on a test set.

        Args:
            dataset: The dataset to evaluate metrics on.
            gt_column: The dataset column to use as ground truth.
            predictions: Model outputs to evaluate.
            model_input_column: (Optional) For autoregistered models,
                the prediction sometimes contains the model input.
                So we need to delete the model input if it's in the predictions.
            metrics: (Optional) The metrics to use, defaults to using
                chr_f, exact_match, and bert_score.

        Returns:
            A dictionary of metric values to return.
        """
        if metrics is not None:
            metric_names = [each.name for each in metrics]
            metric_names = sorted(metric_names, key=lambda name: name.lower())
            assert set(metric_names) < {
                "chr_f",
                "exact_match",
                "bert_score",
            }, "Metrics must be within chr_f, exact_match, and bert_score."
            logger.info(f"Using selected metrics: {', '.join(metric_names)}.")
        else:
            logger.info("Using default metrics of chr_f, exact_match and bert_score.")
            metrics = [
                evaluate.load("chrf"),
                evaluate.load("exact_match"),
                evaluate.load("bertscore"),
            ]
        # Get the ground truth from the dataset
        ground_truths = dataset[gt_column]
        # Extract the predicted strings from ModelOutput
        predicted_strings = [each.prediction for each in predictions]
        assert len(ground_truths) == len(
            predicted_strings
        ), "The length of input dataset and predictions are not equal."
        # Initialize the metric values dictionary
        metric_values = {}

        if model_input_column is not None:
            # Some of the autoregregistered models' output always contains
            # the input. So we need to delete the model input if it's in the
            # predictions when necessary.
            logger.info(
                "The model_input_column is not None. The model input will be detached from predictions if necessary."  # noqa E501
            )
            model_inputs = dataset[model_input_column]
            for idx, model_input in enumerate(model_inputs):
                ground_truth = ground_truths[idx]
                predicted_string = predicted_strings[idx]
                if (model_input in predicted_string) and (
                    model_input not in ground_truth
                ):
                    predicted_string = predicted_string.replace(model_input, "")
                    predicted_strings[idx] = predicted_string

        # Compute and store metric values
        for metric in metrics:
            metric_name = metric.name
            metric.add_batch(predictions=predicted_strings, references=ground_truths)
            if metric_name == "chr_f":
                metric_values["chr_f++"] = metric.compute(word_order=2)["score"]
            elif metric_name == "exact_match":
                metric_values[metric_name] = metric.compute()["exact_match"]
            elif metric_name == "bert_score":
                metric_values["average_bert_score"] = np.average(
                    metric.compute(model_type=encoder_model_name)["f1"]
                )

        return metric_values



"""An interface for model selection."""

from abc import ABC, abstractmethod

from prompt2model.prompt_parser import PromptSpec


# pylint: disable=too-few-public-methods
class ModelRetriever(ABC):
    """Retrieve several models from HuggingFace."""

    @abstractmethod
    def retrieve(
        self,
        prompt: PromptSpec,
    ) -> str:
        """Retrieve relevant models from HuggingFace.

        Args:
            prompt: A prompt to use to select relevant models.

        Return:
            A relevant model's HuggingFace name.
        """



"""Import DatasetRetriever classes."""
from prompt2model.dataset_retriever.base import DatasetRetriever
from prompt2model.dataset_retriever.hf_dataset_retriever import (
    DatasetInfo,
    DescriptionDatasetRetriever,
)
from prompt2model.dataset_retriever.mock import MockRetriever

__all__ = (
    "DatasetRetriever",
    "MockRetriever",
    "DescriptionDatasetRetriever",
    "DatasetInfo",
)



"""A class for generating empty datasets (for testing purposes)."""

import datasets

from prompt2model.dataset_generator.base import DatasetGenerator, DatasetSplit
from prompt2model.prompt_parser import PromptSpec


class MockDatasetGenerator(DatasetGenerator):
    """A class for generating empty datasets (for testing purposes)."""

    def generate_dataset_split(
        self,
        prompt_spec: PromptSpec,
        expected_num_examples: int,
        split: DatasetSplit,
    ) -> datasets.Dataset:
        """Create empty versions of the datasets, for testing.

        Args:
            prompt_spec: A prompt specification.
            expected_num_examples: Number of examples in split.
            split: Name of dataset split to generate.

        Returns:
            A single dataset split.
        """
        _ = prompt_spec, split  # suppress unused variable warnings
        col_values = [""] * expected_num_examples
        # Construct empty-valued dataframe with length matching expected_num_examples.
        return datasets.Dataset.from_dict(
            {"input_col": col_values, "output_col": col_values}
        )



"""An interface for dataset retrieval."""

from __future__ import annotations  # noqa FI58

from abc import ABC, abstractmethod

import datasets

from prompt2model.prompt_parser import PromptSpec


# pylint: disable=too-few-public-methods
class DatasetRetriever(ABC):
    """A class for retrieving datasets."""

    @abstractmethod
    def retrieve_dataset_dict(
        self, prompt_spec: PromptSpec
    ) -> datasets.DatasetDict | None:
        """Retrieve full dataset splits (e.g. train/dev/test) from a prompt.

        Args:
            prompt_spec: A prompt spec (containing a system description).

        Returns:
            A retrieved DatasetDict containing train/val/test splits.
        """



"""An dual-encoder dataset retriever using HuggingFace dataset descriptions."""

from __future__ import annotations  # noqa FI58

import json
import os
import urllib.request

import datasets
import numpy as np
import torch

from prompt2model.dataset_retriever.base import DatasetRetriever
from prompt2model.prompt_parser import PromptSpec
from prompt2model.utils import encode_text, retrieve_objects

datasets.utils.logging.disable_progress_bar()


class DatasetInfo:
    """Store the dataset name, description, and query-dataset score for each dataset."""

    def __init__(
        self,
        name: str,
        description: str,
        score: float,
    ):
        """Initialize a DatasetInfo object.

        Args:
            name: The name of the dataset.
            description: The description of the dataset.
            score: The similarity of the dataset to a given prompt from a user.
        """
        self.name = name
        self.description = description
        self.score = score


def input_string():
    """Read a string from stdin."""
    description = str(input())
    return description


def input_y_n() -> bool:
    """Get a yes/no answer from the user via stdin."""
    y_n = str(input())
    return not (y_n.strip() == "" or y_n.strip().lower() == "n")


class DescriptionDatasetRetriever(DatasetRetriever):
    """Retrieve a dataset from HuggingFace, based on similarity to the prompt."""

    def __init__(
        self,
        search_index_path: str = "huggingface_data/huggingface_datasets/"
        + "huggingface_datasets_datafinder_index",
        first_stage_search_depth: int = 1000,
        max_search_depth: int = 25,
        encoder_model_name: str = "viswavi/datafinder-huggingface-prompt-queries",
        dataset_info_file: str = "huggingface_data/huggingface_datasets/"
        + "dataset_index.json",
        device: torch.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        ),
    ):
        """Initialize a dual-encoder retriever against a search index.

        Args:
            search_index_path: Where to store the search index (e.g. encoded vectors).
            first_stage_search_depth: The # of datasets to retrieve before filtering.
            max_search_depth: The number of most-relevant datasets to retrieve.
            encoder_model_name: The name of the model to use for the dual-encoder.
            dataset_info_file: The file containing dataset names and descriptions.
            device: The device to use for encoding text for our dual-encoder model.
        """
        self.search_index_path = search_index_path
        self.first_stage_search_depth = first_stage_search_depth
        self.max_search_depth = max_search_depth
        self.encoder_model_name = encoder_model_name
        self.dataset_infos: list[DatasetInfo] = []
        if not os.path.exists(dataset_info_file):
            # Download the dataset search index if one is not on disk already.
            urllib.request.urlretrieve(
                "http://phontron.com/data/prompt2model/dataset_index.json",
                dataset_info_file,
            )
        self.full_dataset_metadata = json.load(open(dataset_info_file, "r"))
        for dataset_name in sorted(self.full_dataset_metadata.keys()):
            self.dataset_infos.append(
                DatasetInfo(
                    name=dataset_name,
                    description=self.full_dataset_metadata[dataset_name]["description"],
                    score=0.0,
                )
            )
        self.device = device

        assert not os.path.isdir(
            search_index_path
        ), f"Search index must either be a valid file or not exist yet. But {search_index_path} is provided."  # noqa E501

    def encode_dataset_descriptions(
        self, dataset_infos, search_index_path
    ) -> np.ndarray:
        """Encode dataset descriptions into a vector for indexing."""
        dataset_descriptions = [
            dataset_info.description for dataset_info in dataset_infos
        ]
        dataset_vectors = encode_text(
            self.encoder_model_name,
            text_to_encode=dataset_descriptions,
            encoding_file=search_index_path,
            device=self.device,
        )
        return dataset_vectors

    @staticmethod
    def print_divider():
        """Utility function to assist with the retriever's command line interface."""
        print("\n-------------------------------------------------\n")

    def choose_dataset(self, top_datasets: list[DatasetInfo]) -> str | None:
        """Have the user choose an appropriate dataset from a list of top datasets."""
        self.print_divider()
        print("Here are the datasets I've retrieved for you:")
        print("#\tName\tDescription")
        for i, d in enumerate(top_datasets):
            description_no_spaces = d.description.replace("\n", " ")
            print(f"{i+1}):\t{d.name}\t{description_no_spaces}")

        self.print_divider()
        print(
            "If none of these are relevant to your prompt, we'll only use "
            + "generated data. Are any of these datasets relevant? (y/N)"
        )
        any_datasets_relevant = input_y_n()
        if any_datasets_relevant:
            print(
                "Which dataset would you like to use? Give the number between "
                + f"1 and {len(top_datasets)}."
            )
            dataset_idx = int(input())
            chosen_dataset_name = top_datasets[dataset_idx - 1].name
        else:
            chosen_dataset_name = None
        self.print_divider()  # noqa E501
        return chosen_dataset_name

    @staticmethod
    def canonicalize_dataset_using_columns_for_split(
        dataset_split: datasets.Dataset,
        input_columns: list[str],
        output_column: str,
    ) -> datasets.DatasetDict:
        """Canonicalize a single dataset split into a suitable text-to-text format."""
        input_col = []
        output_col = []
        for i in range(len(dataset_split)):
            input_string = ""
            for col in input_columns:
                input_string += f"{col}: {dataset_split[i][col]}\n"
            input_string = input_string.strip()
            input_col.append(input_string)
            output_col.append(dataset_split[i][output_column])
        return datasets.Dataset.from_dict(
            {"input_col": input_col, "output_col": output_col}
        )

    def canonicalize_dataset_using_columns(
        self,
        dataset: datasets.DatasetDict,
        input_columns: list[str],
        output_columns: str,
    ) -> datasets.DatasetDict:
        """Canonicalize a dataset into a suitable text-to-text format."""
        dataset_dict = {}
        for split in dataset:
            dataset_dict[split] = self.canonicalize_dataset_using_columns_for_split(
                dataset[split], input_columns, output_columns
            )
        return datasets.DatasetDict(dataset_dict)

    def canonicalize_dataset(self, dataset_name: str) -> datasets.DatasetDict:
        """Canonicalize a dataset into a suitable text-to-text format."""
        configs = datasets.get_dataset_config_names(dataset_name)
        chosen_config = None
        if len(configs) == 1:
            chosen_config = configs[0]
        else:
            self.print_divider()
            print(f"Multiple dataset configs available: {configs}")
            while chosen_config is None:
                print("Which dataset config would you like to use for this?")
                user_response = input_string()
                if user_response in configs:
                    chosen_config = user_response
                else:
                    print(
                        f"Invalid config provided: {user_response}. Please choose "
                        + "from {configs}\n\n"
                    )
            self.print_divider()

        dataset = datasets.load_dataset(dataset_name, chosen_config)
        assert "train" in dataset
        train_columns = dataset["train"].column_names
        train_columns_formatted = ", ".join(train_columns)

        assert len(dataset["train"]) > 0
        example_rows = json.dumps(dataset["train"][0], indent=4)

        self.print_divider()
        print(f"Loaded dataset. Example row:\n{example_rows}\n")

        print(
            "Which column(s) should we use as input? Provide a comma-separated "
            + f"list from: {train_columns_formatted}."
        )
        user_response = input_string()
        input_columns = [c.strip() for c in user_response.split(",")]
        print(f"Will use the columns {json.dumps(input_columns)} as input.\n")

        output_column = None
        while output_column is None:
            print(
                "Which column(s) should we use as the target? Choose a single "
                + f"value from: {train_columns_formatted}."
            )
            user_response = input_string()
            if user_response in train_columns:
                output_column = user_response
            else:
                print(
                    "Invalid column provided: {user_response}. Please choose "
                    + f"from {train_columns}\n\n"
                )
        print(f'Will use the column "{output_column}" as our target.\n')
        self.print_divider()

        canonicalized_dataset = self.canonicalize_dataset_using_columns(
            dataset, input_columns, output_column
        )
        return canonicalized_dataset

    def retrieve_dataset_dict(
        self,
        prompt_spec: PromptSpec,
    ) -> datasets.DatasetDict | None:
        """Select a dataset from a prompt using a dual-encoder retriever.

        Args:
            prompt_spec: A prompt whose instruction field we use to retrieve datasets.

        Return:
            A list of relevant datasets dictionaries.
        """
        if not os.path.exists(self.search_index_path):
            print("Creating dataset descriptions")
            self.encode_dataset_descriptions(self.dataset_infos, self.search_index_path)

        query_text = prompt_spec.instruction

        query_vector = encode_text(
            self.encoder_model_name,
            text_to_encode=query_text,
            device=self.device,
        )
        dataset_names = [dataset_info.name for dataset_info in self.dataset_infos]
        ranked_list = retrieve_objects(
            query_vector,
            self.search_index_path,
            dataset_names,
            self.first_stage_search_depth,
        )
        top_dataset_infos = []
        dataset_name_to_dataset_idx = {
            d.name: i for i, d in enumerate(self.dataset_infos)
        }
        for dataset_name, dataset_score in ranked_list:
            dataset_idx = dataset_name_to_dataset_idx[dataset_name]
            self.dataset_infos[dataset_idx].score = dataset_score
            top_dataset_infos.append(self.dataset_infos[dataset_idx])

        ranked_list = sorted(top_dataset_infos, key=lambda x: x.score, reverse=True)[
            : self.max_search_depth
        ]
        assert len(ranked_list) > 0, "No datasets retrieved from search index."
        top_dataset_name = self.choose_dataset(ranked_list)
        if top_dataset_name is None:
            return None
        return self.canonicalize_dataset(top_dataset_name)



"""Import DatasetGenerator classes."""
from prompt2model.dataset_generator.base import DatasetGenerator, DatasetSplit
from prompt2model.dataset_generator.mock import MockDatasetGenerator
from prompt2model.dataset_generator.openai_gpt import OpenAIDatasetGenerator

__all__ = (
    "OpenAIDatasetGenerator",
    "MockDatasetGenerator",
    "DatasetGenerator",
    "DatasetSplit",
)



"""An interface for creating Gradio demos automatically."""

import gradio as gr
import transformers

from prompt2model.prompt_parser.base import PromptSpec


def mock_gradio_create(
    model: transformers.PreTrainedModel, prompt_spec: PromptSpec
) -> gr.Interface:
    """Create a Gradio interface automatically.

    Args:
        model: A trained model to expose via a Gradio interface.
        prompt_spec: A PromptSpec to help choose the visual interface.

    Returns:
        A Gradio interface for interacting with the model.

    """
    _ = model, prompt_spec  # suppress unused variable warnings
    dummy_interface = gr.Interface(lambda input: None, "textbox", "label")
    return dummy_interface



"""An interface for prompt parsing."""

from __future__ import annotations  # noqa FI58

import json
import os

import openai

from prompt2model.prompt_parser.base import PromptSpec, TaskType

from prompt2model.prompt_parser.instr_parser_prompt import (  # isort: split
    construct_prompt_for_instruction_parsing,
)
from prompt2model.utils import (
    OPENAI_ERRORS,
    ChatGPTAgent,
    get_formatted_logger,
    handle_openai_error,
)

logger = get_formatted_logger("PromptParser")

os.environ["TOKENIZERS_PARALLELISM"] = "false"


class OpenAIInstructionParser(PromptSpec):
    """Parse the prompt to separate instructions from task demonstrations."""

    def __init__(
        self, task_type: TaskType, api_key: str | None = None, max_api_calls: int = None
    ):
        """Initialize the prompt spec with empty parsed fields.

        We initialize the "instruction" and "examples" fields with None.
        These fields can be populated with the parse_from_prompt method.

        Args:
            task_type: Set a constant task type to use for all prompts.
            api_key: A valid OpenAI API key. Alternatively, set as None and set
                the environment variable with `export OPENAI_API_KEY=<your key>`.
            max_api_calls: The maximum number of API calls allowed,
                or None for unlimited.
        """
        self.task_type = task_type
        self._instruction: str | None = None
        self._examples: str | None = None
        self.api_key: str | None = api_key if api_key else os.environ["OPENAI_API_KEY"]
        assert self.api_key is not None and self.api_key != "", (
            "API key must be provided"
            + " or set the environment variable with `export OPENAI_API_KEY=<your key>`"
        )
        if max_api_calls:
            assert max_api_calls > 0, "max_api_calls must be > 0"
        self.max_api_calls = max_api_calls
        self.api_call_counter = 0

    def extract_response(self, response: openai.Completion) -> tuple[str, str] | None:
        """Parse stuctured fields from the OpenAI API response.

        Args:
            response: OpenAI API response.

        Returns:
            If the API response is a valid JSON object and contains the required_keys,
                then returns a tuple consisting of:
                1) Instruction: The instruction parsed from the API response.
                2) Demonstrations: (Optional) demonstrations parsed from the
                API response.
            Else returns None.
        """
        response_text = response.choices[0]["message"]["content"]
        try:
            response_json = json.loads(response_text, strict=False)
        except json.decoder.JSONDecodeError:
            logger.warning("API response was not a valid JSON")
            return None

        required_keys = ["Instruction", "Demonstrations"]
        missing_keys = [key for key in required_keys if key not in response_json]
        if len(missing_keys) != 0:
            logger.warning(f'API response must contain {", ".join(required_keys)} keys')
            return None
        instruction_string = response_json["Instruction"].strip()
        demonstration_string = response_json["Demonstrations"].strip()
        return instruction_string, demonstration_string

    def parse_from_prompt(self, prompt: str) -> None:
        """Parse prompt into specific fields, stored as class member variables.

        This function directly stores the parsed fields into the class's member
        variables `instruction` and `examples`. So it has no return value.

        Args:
            prompt: User prompt to parse into two specific fields:
                    "instruction" and "demonstrations".
        """
        parsing_prompt_for_chatgpt = construct_prompt_for_instruction_parsing(prompt)

        chat_api = ChatGPTAgent(self.api_key)
        while True:
            try:
                self.api_call_counter += 1
                response = chat_api.generate_one_openai_chat_completion(
                    parsing_prompt_for_chatgpt,
                    temperature=0,
                    presence_penalty=0,
                    frequency_penalty=0,
                )
                extraction = self.extract_response(response)
                if extraction is not None:
                    self._instruction, self._examples = extraction
                    return None
                else:
                    if (
                        self.max_api_calls
                        and self.api_call_counter == self.max_api_calls
                    ):
                        logger.warning(
                            "Maximum number of API calls reached for PromptParser."
                        )
                        return None
            except OPENAI_ERRORS as e:
                self.api_call_counter = handle_openai_error(e, self.api_call_counter)
                if self.max_api_calls and self.api_call_counter >= self.max_api_calls:
                    logger.error("Maximum number of API calls reached.")
                    raise ValueError("Maximum number of API calls reached.") from e



"""Tools for generating hypothetical documents from prompts."""

from __future__ import annotations  # noqa FI58

import logging

from prompt2model.prompt_parser import PromptSpec
from prompt2model.utils import OPENAI_ERRORS, ChatGPTAgent, handle_openai_error

PROMPT_PREFIX = '''HuggingFace contains models, which are each given a user-generated description. The first section of the description, delimited with two "---" lines, consists of a YAML description of the model. This may contain fields like "language" (supported by model), "datasets" (used to train the model), "tags" (e.g. tasks relevant to the model), and "metrics" (used to evaluate the model). Create a hypothetical HuggingFace model description that would satisfy a given user instruction. Here are some examples:

Instruction: "Give me some translation from English to Vietnamese. Input English and output Vietnamese."
Hypothetical model description:
---
language:
- en
- vi

tags:
- translation

license: apache-2.0
---

### eng-vie

* source group: English
* target group: Vietnamese
*  OPUS readme: [eng-vie](https://github.com/Helsinki-NLP/Tatoeba-Challenge/tree/master/models/eng-vie/README.md)

*  model: transformer-align
* source language(s): eng
* target language(s): vie vie_Hani
* model: transformer-align
* pre-processing: normalization + SentencePiece (spm32k,spm32k)
* a sentence initial language token is required in the form of `>>id<<` (id = valid target language ID)
* download original weights: [opus-2020-06-17.zip](https://object.pouta.csc.fi/Tatoeba-MT-models/eng-vie/opus-2020-06-17.zip)
* test set translations: [opus-2020-06-17.test.txt](https://object.pouta.csc.fi/Tatoeba-MT-models/eng-vie/opus-2020-06-17.test.txt)
* test set scores: [opus-2020-06-17.eval.txt](https://object.pouta.csc.fi/Tatoeba-MT-models/eng-vie/opus-2020-06-17.eval.txt)

## Benchmarks

| testset               | BLEU  | chr-F |
|-----------------------|-------|-------|
| Tatoeba-test.eng.vie 	| 37.2 	| 0.542 |


### System Info:
- hf_name: eng-vie

- source_languages: eng

- target_languages: vie

- opus_readme_url: https://github.com/Helsinki-NLP/Tatoeba-Challenge/tree/master/models/eng-vie/README.md

- original_repo: Tatoeba-Challenge

- tags: ['translation']

- languages: ['en', 'vi']

- src_constituents: {'eng'}

- tgt_constituents: {'vie', 'vie_Hani'}

- src_multilingual: False

- tgt_multilingual: False

- prepro:  normalization + SentencePiece (spm32k,spm32k)

- url_model: https://object.pouta.csc.fi/Tatoeba-MT-models/eng-vie/opus-2020-06-17.zip

- url_test_set: https://object.pouta.csc.fi/Tatoeba-MT-models/eng-vie/opus-2020-06-17.test.txt

- src_alpha3: eng

- tgt_alpha3: vie

- short_pair: en-vi

- chrF2_score: 0.542

- bleu: 37.2

- brevity_penalty: 0.973

- ref_len: 24427.0

- src_name: English

- tgt_name: Vietnamese

- train_date: 2020-06-17

- src_alpha2: en

- tgt_alpha2: vi

- prefer_old: False

- long_pair: eng-vie

- helsinki_git_sha: 480fcbe0ee1bf4774bcbe6226ad9f58e63f6c535

- transformers_git_sha: 2207e5d8cb224e954a7cba69fa4ac2309e9ff30b

- port_machine: brutasse

- port_time: 2020-08-21-14:41


Instruction: "I want to summarize things like news articles."
Hypothetical model description:
---
language: en
license: apache-2.0
tags:
- pegasus
- seq2seq
- summarization
model-index:
- name: tuner007/pegasus_summarizer
  results:
  - task:
      type: summarization
      name: Summarization
    dataset:
      name: cnn_dailymail
      type: cnn_dailymail
      config: 3.0.0
      split: train
    metrics:
    - name: ROUGE-1
      type: rouge
      value: 36.604
      verified: true
    - name: ROUGE-2
      type: rouge
      value: 14.6398
      verified: true
    - name: ROUGE-L
      type: rouge
      value: 23.8845
      verified: true
    - name: ROUGE-LSUM
      type: rouge
      value: 32.9017
      verified: true
    - name: loss
      type: loss
      value: 2.5757133960723877
      verified: true
    - name: gen_len
      type: gen_len
      value: 76.3984
      verified: true
---

## Model description
[PEGASUS](https://github.com/google-research/pegasus) fine-tuned for summarization

## Install "sentencepiece" library required for tokenizer
```
pip install sentencepiece
```

## Model in Action 🚀
```
import torch
from transformers import PegasusForConditionalGeneration, PegasusTokenizer
model_name = 'tuner007/pegasus_summarizer'
torch_device = 'cuda' if torch.cuda.is_available() else 'cpu'
tokenizer = PegasusTokenizer.from_pretrained(model_name)
model = PegasusForConditionalGeneration.from_pretrained(model_name).to(torch_device)

def get_response(input_text):
  batch = tokenizer([input_text],truncation=True,padding='longest',max_length=1024, return_tensors="pt").to(torch_device)
  gen_out = model.generate(**batch,max_length=128,num_beams=5, num_return_sequences=1, temperature=1.5)
  output_text = tokenizer.batch_decode(gen_out, skip_special_tokens=True)
  return output_text
```
#### Example:
context = """"
India wicket-keeper batsman Rishabh Pant has said someone from the crowd threw a ball on pacer Mohammed Siraj while he was fielding in the ongoing third Test against England on Wednesday. Pant revealed the incident made India skipper Virat Kohli "upset". "I think, somebody threw a ball inside, at Siraj, so he [Kohli] was upset," said Pant in a virtual press conference after the close of the first day\'s play."You can say whatever you want to chant, but don\'t throw things at the fielders and all those things. It is not good for cricket, I guess," he added.In the third session of the opening day of the third Test, a section of spectators seemed to have asked Siraj the score of the match to tease the pacer. The India pacer however came with a brilliant reply as he gestured 1-0 (India leading the Test series) towards the crowd.Earlier this month, during the second Test match, there was some bad crowd behaviour on a show as some unruly fans threw champagne corks at India batsman KL Rahul.Kohli also intervened and he was seen gesturing towards the opening batsman to know more about the incident. An over later, the TV visuals showed that many champagne corks were thrown inside the playing field, and the Indian players were visibly left frustrated.Coming back to the game, after bundling out India for 78, openers Rory Burns and Haseeb Hameed ensured that England took the honours on the opening day of the ongoing third Test.At stumps, England\'s score reads 120/0 and the hosts have extended their lead to 42 runs. For the Three Lions, Burns (52*) and Hameed (60*) are currently unbeaten at the crease.Talking about the pitch on opening day, Pant said, "They took the heavy roller, the wicket was much more settled down, and they batted nicely also," he said. "But when we batted, the wicket was slightly soft, and they bowled in good areas, but we could have applied [ourselves] much better."Both England batsmen managed to see off the final session and the hosts concluded the opening day with all ten wickets intact, extending the lead to 42.(ANI)
"""

```
get_response(context)
```
#### Output:
Team India wicketkeeper-batsman Rishabh Pant has said that Virat Kohli was "upset" after someone threw a ball on pacer Mohammed Siraj while he was fielding in the ongoing third Test against England. "You can say whatever you want to chant, but don't throw things at the fielders and all those things. It's not good for cricket, I guess," Pant added.'

#### [Inshort](https://www.inshorts.com/) (60 words News summary app, rated 4.4 by 5,27,246+ users on android playstore) summary:
India wicketkeeper-batsman Rishabh Pant has revealed that captain Virat Kohli was upset with the crowd during the first day of Leeds Test against England because someone threw a ball at pacer Mohammed Siraj. Pant added, "You can say whatever you want to chant, but don't throw things at the fielders and all those things. It is not good for cricket."


> Created by [Arpit Rajauria](https://twitter.com/arpit_rajauria)
[![Twitter icon](https://cdn0.iconfinder.com/data/icons/shift-logotypes/32/Twitter-32.png)](https://twitter.com/arpit_rajauria)


### Framework versions

- Transformers 4.31.0
- Pytorch 2.0.1+cu118
- Datasets 2.13.1
- Tokenizers 0.13.3


Instruction: "I want to classify sentences by their sentiment (positive/negative/neutral)."
Hypothetical model description:
---
language: en
license: apache-2.0
datasets:
- sst2
- glue
model-index:
- name: distilbert-base-uncased-finetuned-sst-2-english
  results:
  - task:
      type: text-classification
      name: Text Classification
    dataset:
      name: glue
      type: glue
      config: sst2
      split: validation
    metrics:
    - type: accuracy
      value: 0.9105504587155964
      name: Accuracy
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiN2YyOGMxYjY2Y2JhMjkxNjIzN2FmMjNiNmM2ZWViNGY3MTNmNWI2YzhiYjYxZTY0ZGUyN2M1NGIxZjRiMjQwZiIsInZlcnNpb24iOjF9.uui0srxV5ZHRhxbYN6082EZdwpnBgubPJ5R2-Wk8HTWqmxYE3QHidevR9LLAhidqGw6Ih93fK0goAXncld_gBg
    - type: precision
      value: 0.8978260869565218
      name: Precision
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiMzgwYTYwYjA2MmM0ZTYwNDk0M2NmNTBkZmM2NGNhYzQ1OGEyN2NkNDQ3Mzc2NTQyMmZiNDJiNzBhNGVhZGUyOSIsInZlcnNpb24iOjF9.eHjLmw3K02OU69R2Au8eyuSqT3aBDHgZCn8jSzE3_urD6EUSSsLxUpiAYR4BGLD_U6-ZKcdxVo_A2rdXqvUJDA
    - type: recall
      value: 0.9301801801801802
      name: Recall
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiMGIzM2E3MTI2Mzc2MDYwNmU3ZTVjYmZmZDBkNjY4ZTc5MGY0Y2FkNDU3NjY1MmVkNmE3Y2QzMzAwZDZhOWY1NiIsInZlcnNpb24iOjF9.PUZlqmct13-rJWBXdHm5tdkXgETL9F82GNbbSR4hI8MB-v39KrK59cqzFC2Ac7kJe_DtOeUyosj34O_mFt_1DQ
    - type: auc
      value: 0.9716626673402374
      name: AUC
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiMDM0YWIwZmQ4YjUwOGZmMWU2MjI1YjIxZGQ2MzNjMzRmZmYxMzZkNGFjODhlMDcyZDM1Y2RkMWZlOWQ0MWYwNSIsInZlcnNpb24iOjF9.E7GRlAXmmpEkTHlXheVkuL1W4WNjv4JO3qY_WCVsTVKiO7bUu0UVjPIyQ6g-J1OxsfqZmW3Leli1wY8vPBNNCQ
    - type: f1
      value: 0.9137168141592922
      name: F1
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiMGU4MjNmOGYwZjZjMDQ1ZTkyZTA4YTc1MWYwOTM0NDM4ZWY1ZGVkNDY5MzNhYTQyZGFlNzIyZmUwMDg3NDU0NyIsInZlcnNpb24iOjF9.mW5ftkq50Se58M-jm6a2Pu93QeKa3MfV7xcBwvG3PSB_KNJxZWTCpfMQp-Cmx_EMlmI2siKOyd8akYjJUrzJCA
    - type: loss
      value: 0.39013850688934326
      name: loss
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiMTZiNzAyZDc0MzUzMmE1MGJiN2JlYzFiODE5ZTNlNGE4MmI4YzRiMTc2ODEzMTUwZmEzOTgxNzc4YjJjZTRmNiIsInZlcnNpb24iOjF9.VqIC7uYC-ZZ8ss9zQOlRV39YVOOLc5R36sIzCcVz8lolh61ux_5djm2XjpP6ARc6KqEnXC4ZtfNXsX2HZfrtCQ
  - task:
      type: text-classification
      name: Text Classification
    dataset:
      name: sst2
      type: sst2
      config: default
      split: train
    metrics:
    - type: accuracy
      value: 0.9885521685548412
      name: Accuracy
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiY2I3NzU3YzhmMDkxZTViY2M3OTY1NmI0ZTdmMDQxNjNjYzJiZmQxNzczM2E4YmExYTY5ODY0NDBkY2I4ZjNkOCIsInZlcnNpb24iOjF9.4Gtk3FeVc9sPWSqZIaeUXJ9oVlPzm-NmujnWpK2y5s1Vhp1l6Y1pK5_78wW0-NxSvQqV6qd5KQf_OAEpVAkQDA
    - type: precision
      value: 0.9881965062029833
      name: Precision Macro
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiZDdlZDMzY2I3MTAwYTljNmM4MGMyMzU2YjAzZDg1NDYwN2ZmM2Y5OWZhMjUyMGJiNjY1YmZiMzFhMDI2ODFhNyIsInZlcnNpb24iOjF9.cqmv6yBxu4St2mykRWrZ07tDsiSLdtLTz2hbqQ7Gm1rMzq9tdlkZ8MyJRxtME_Y8UaOG9rs68pV-gKVUs8wABw
    - type: precision
      value: 0.9885521685548412
      name: Precision Micro
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiZjFlYzAzNmE1YjljNjUwNzBjZjEzZDY0ZDQyMmY5ZWM2OTBhNzNjYjYzYTk1YWE1NjU3YTMxZDQwOTE1Y2FkNyIsInZlcnNpb24iOjF9.jnCHOkUHuAOZZ_ZMVOnetx__OVJCS6LOno4caWECAmfrUaIPnPNV9iJ6izRO3sqkHRmxYpWBb-27GJ4N3LU-BQ
    - type: precision
      value: 0.9885639626373408
      name: Precision Weighted
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiZGUyODFjNjBlNTE2MTY3ZDAxOGU1N2U0YjUyY2NiZjhkOGVmYThjYjBkNGU3NTRkYzkzNDQ2MmMwMjkwMWNiMyIsInZlcnNpb24iOjF9.zTNabMwApiZyXdr76QUn7WgGB7D7lP-iqS3bn35piqVTNsv3wnKjZOaKFVLIUvtBXq4gKw7N2oWxvWc4OcSNDg
    - type: recall
      value: 0.9886145346602994
      name: Recall Macro
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiNTU1YjlhODU3YTkyNTdiZDcwZGFlZDBiYjY0N2NjMGM2NTRiNjQ3MDNjNGMxOWY2ZGQ4NWU1YmMzY2UwZTI3YSIsInZlcnNpb24iOjF9.xaLPY7U-wHsJ3DDui1yyyM-xWjL0Jz5puRThy7fczal9x05eKEQ9s0a_WD-iLmapvJs0caXpV70hDe2NLcs-DA
    - type: recall
      value: 0.9885521685548412
      name: Recall Micro
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiODE0YTU0MDBlOGY4YzU0MjY5MzA3OTk2OGNhOGVkMmU5OGRjZmFiZWI2ZjY5ODEzZTQzMTI0N2NiOTVkNDliYiIsInZlcnNpb24iOjF9.SOt1baTBbuZRrsvGcak2sUwoTrQzmNCbyV2m1_yjGsU48SBH0NcKXicidNBSnJ6ihM5jf_Lv_B5_eOBkLfNWDQ
    - type: recall
      value: 0.9885521685548412
      name: Recall Weighted
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiZWNkNmM0ZGRlNmYxYzIwNDk4OTI5MzIwZWU1NzZjZDVhMDcyNDFlMjBhNDQxODU5OWMwMWNhNGEzNjY3ZGUyOSIsInZlcnNpb24iOjF9.b15Fh70GwtlG3cSqPW-8VEZT2oy0CtgvgEOtWiYonOovjkIQ4RSLFVzVG-YfslaIyfg9RzMWzjhLnMY7Bpn2Aw
    - type: f1
      value: 0.9884019815052447
      name: F1 Macro
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiYmM4NjQ5Yjk5ODRhYTU1MTY3MmRhZDBmODM1NTg3OTFiNWM4NDRmYjI0MzZkNmQ1MzE3MzcxODZlYzBkYTMyYSIsInZlcnNpb24iOjF9.74RaDK8nBVuGRl2Se_-hwQvP6c4lvVxGHpcCWB4uZUCf2_HoC9NT9u7P3pMJfH_tK2cpV7U3VWGgSDhQDi-UBQ
    - type: f1
      value: 0.9885521685548412
      name: F1 Micro
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiZDRmYWRmMmQ0YjViZmQxMzhhYTUyOTE1MTc0ZDU1ZjQyZjFhMDYzYzMzZDE0NzZlYzQyOTBhMTBhNmM5NTlkMiIsInZlcnNpb24iOjF9.VMn_psdAHIZTlW6GbjERZDe8MHhwzJ0rbjV_VJyuMrsdOh5QDmko-wEvaBWNEdT0cEKsbggm-6jd3Gh81PfHAQ
    - type: f1
      value: 0.9885546181087554
      name: F1 Weighted
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiMjUyZWFhZDZhMGQ3MzBmYmRiNDVmN2FkZDBjMjk3ODk0OTAxNGZkMWE0NzU5ZjI0NzE0NGZiNzM0N2Y2NDYyOSIsInZlcnNpb24iOjF9.YsXBhnzEEFEW6jw3mQlFUuIrW7Gabad2Ils-iunYJr-myg0heF8NEnEWABKFE1SnvCWt-69jkLza6SupeyLVCA
    - type: loss
      value: 0.040652573108673096
      name: loss
      verified: true
      verifyToken: eyJhbGciOiJFZERTQSIsInR5cCI6IkpXVCJ9.eyJoYXNoIjoiZTc3YjU3MjdjMzkxODA5MjU5NGUyY2NkMGVhZDg3ZWEzMmU1YWVjMmI0NmU2OWEyZTkzMTVjNDZiYTc0YjIyNCIsInZlcnNpb24iOjF9.lA90qXZVYiILHMFlr6t6H81Oe8a-4KmeX-vyCC1BDia2ofudegv6Vb46-4RzmbtuKeV6yy6YNNXxXxqVak1pAg
---

# DistilBERT base uncased finetuned SST-2

## Table of Contents
- [Model Details](#model-details)
- [How to Get Started With the Model](#how-to-get-started-with-the-model)
- [Uses](#uses)
- [Risks, Limitations and Biases](#risks-limitations-and-biases)
- [Training](#training)

## Model Details
**Model Description:** This model is a fine-tune checkpoint of [DistilBERT-base-uncased](https://huggingface.co/distilbert-base-uncased), fine-tuned on SST-2.
This model reaches an accuracy of 91.3 on the dev set (for comparison, Bert bert-base-uncased version reaches an accuracy of 92.7).
- **Developed by:** Hugging Face
- **Model Type:** Text Classification
- **Language(s):** English
- **License:** Apache-2.0
- **Parent Model:** For more details about DistilBERT, we encourage users to check out [this model card](https://huggingface.co/distilbert-base-uncased).
- **Resources for more information:**
    - [Model Documentation](https://huggingface.co/docs/transformers/main/en/model_doc/distilbert#transformers.DistilBertForSequenceClassification)
    - [DistilBERT paper](https://arxiv.org/abs/1910.01108)

## How to Get Started With the Model

Example of single-label classification:
​​
```python
import torch
from transformers import DistilBertTokenizer, DistilBertForSequenceClassification

tokenizer = DistilBertTokenizer.from_pretrained("distilbert-base-uncased")
model = DistilBertForSequenceClassification.from_pretrained("distilbert-base-uncased")

inputs = tokenizer("Hello, my dog is cute", return_tensors="pt")
with torch.no_grad():
    logits = model(**inputs).logits

predicted_class_id = logits.argmax().item()
model.config.id2label[predicted_class_id]

```

## Uses

#### Direct Use

This model can be used for  topic classification. You can use the raw model for either masked language modeling or next sentence prediction, but it's mostly intended to be fine-tuned on a downstream task. See the model hub to look for fine-tuned versions on a task that interests you.

#### Misuse and Out-of-scope Use
The model should not be used to intentionally create hostile or alienating environments for people. In addition, the model was not trained to be factual or true representations of people or events, and therefore using the model to generate such content is out-of-scope for the abilities of this model.


## Risks, Limitations and Biases

Based on a few experimentations, we observed that this model could produce biased predictions that target underrepresented populations.

For instance, for sentences like `This film was filmed in COUNTRY`, this binary classification model will give radically different probabilities for the positive label depending on the country (0.89 if the country is France, but 0.08 if the country is Afghanistan) when nothing in the input indicates such a strong semantic shift. In this [colab](https://colab.research.google.com/gist/ageron/fb2f64fb145b4bc7c49efc97e5f114d3/biasmap.ipynb), [Aurélien Géron](https://twitter.com/aureliengeron) made an interesting map plotting these probabilities for each country.

<img src="https://huggingface.co/distilbert-base-uncased-finetuned-sst-2-english/resolve/main/map.jpeg" alt="Map of positive probabilities per country." width="500"/>

We strongly advise users to thoroughly probe these aspects on their use-cases in order to evaluate the risks of this model. We recommend looking at the following bias evaluation datasets as a place to start: [WinoBias](https://huggingface.co/datasets/wino_bias), [WinoGender](https://huggingface.co/datasets/super_glue), [Stereoset](https://huggingface.co/datasets/stereoset).



# Training


#### Training Data


The authors use the following Stanford Sentiment Treebank([sst2](https://huggingface.co/datasets/sst2)) corpora for the model.

#### Training Procedure

###### Fine-tuning hyper-parameters


- learning_rate = 1e-5
- batch_size = 32
- warmup = 600
- max_seq_length = 128
- num_train_epochs = 3.0
```
:'''  # noqa: E501


def generate_hypothetical_model_description(
    prompt: PromptSpec, openai_api_key: str | None = None, max_api_calls: int = None
) -> str:
    """Generate a hypothetical model description for the user's instruction.

    This method is based on HyDE by Gao et al 2022 (https://arxiv.org/abs/2212.10496).

    Args:
        prompt: PromptSpec object containing the user's instruction.
        openai_api_key: OpenAI API key. If None, use the OPENAI_API_KEY environment
            variable.

    Returns:
        a hypothetical model description for the user's instruction.
    """
    if max_api_calls:
        assert max_api_calls > 0, "max_api_calls must be > 0"
    api_call_counter = 0

    instruction = prompt.instruction
    openai_api_agent = ChatGPTAgent(openai_api_key, "gpt-3.5-turbo-16k")
    chatgpt_prompt = (
        PROMPT_PREFIX
        + "\n"
        + f'Instruction: "{instruction}"\nHypothetical model description:\n'
    )
    while True:
        try:
            chatgpt_completion = openai_api_agent.generate_one_openai_chat_completion(
                chatgpt_prompt,
                temperature=0.0,
                presence_penalty=0.0,
                frequency_penalty=0.0,
            )
            return chatgpt_completion.choices[0]["message"]["content"]
        except OPENAI_ERRORS as e:
            api_call_counter = handle_openai_error(e, api_call_counter)
            if max_api_calls and api_call_counter >= max_api_calls:
                logging.error("Maximum number of API calls reached.")
                raise ValueError("Maximum number of API calls reached.") from e



"""Create a Gradio interface automatically."""

import gradio as gr
import mdtex2html

from prompt2model.model_executor import GenerationModelExecutor
from prompt2model.prompt_parser import OpenAIInstructionParser


def create_gradio(
    model_executor: GenerationModelExecutor, prompt_parser: OpenAIInstructionParser
) -> gr.Blocks:
    """Create a Gradio interface automatically.

    Args:
        model_executor: A GenerationModelExecutor to expose via a Gradio interface.
        prompt_parser: An instance of OpenAIInstructionParser to parse the prompt.

    Returns:
        A Gradio interface for interacting with the model.

    """
    description = prompt_parser.instruction
    examples = prompt_parser.examples

    def postprocess(self, y):
        if y is None:
            return []
        for i, (message, response) in enumerate(y):
            y[i] = (
                None if message is None else mdtex2html.convert((message)),
                None if response is None else mdtex2html.convert(response),
            )
        return y

    gr.Chatbot.postprocess = postprocess

    def response(message):
        response = model_executor.make_single_prediction(message)
        prediction = response.prediction
        return prediction

    def chat(message, history):
        history = history or []
        model_output = (
            response(message) if message != "" else "Please give valid input."
        )
        history.append((message, model_output))
        return history, history

    def reset_user_input():
        return gr.update(value="")

    def reset_state():
        return [], []

    with gr.Blocks() as demo:
        gr.HTML("""<h1 align="center">Prompt2Model</h1>""")
        gr.HTML(f"""<h2 align="center">Task Description: {description}</h2>""")
        gr.HTML(f"""<h2 align="center">Few-shot Examples: {examples}</h2>""")

        chatbot = gr.Chatbot()
        with gr.Row():
            with gr.Column(scale=4):
                with gr.Column(scale=12):
                    user_input = gr.Textbox(
                        show_label=False, placeholder="Input...", lines=10
                    ).style(container=False)
                with gr.Column(min_width=32, scale=2):
                    submitBtn = gr.Button("Submit", variant="primary")
                    emptyBtn = gr.Button("Clear History")

        history = gr.State([])

        submitBtn.click(
            chat, [user_input, history], [chatbot, history], show_progress=True
        )
        submitBtn.click(reset_user_input, [], [user_input])
        emptyBtn.click(reset_state, outputs=[chatbot, history], show_progress=True)

    return demo



"""The real evaluation will be conduted after each mock evaluation of Trainer."""


from transformers import TrainerCallback

from prompt2model.model_evaluator import Seq2SeqEvaluator
from prompt2model.model_executor import GenerationModelExecutor
from prompt2model.utils import get_formatted_logger

logger = get_formatted_logger("ModelTrainer")


class ValidationCallback(TrainerCallback):
    """The real evaluation will be conduted after each mock evaluation of Trainer."""

    def __init__(
        self,
        trainer,
        tokenizer,
        val_dataset,
        executor_batch_size=10,
        tokenizer_max_length=256,
        sequence_max_length=512,
    ) -> None:
        """Initializes a new instance of Model Trainer Callback.

        Args:
            trainer: Trainer instance.
                After each epoch of Training, this callback will be called.
            tokenizer: Tokenizer to initialize model executor.
            val_dataset: Validation dataset to be evaluated on.
            executor_batch_size: The batch size for model executor to
                make predictions.
            tokenizer_max_length: The maximum number of tokens that
                tokenizer is allowed to generate.
            sequence_max_length: The maximum number of tokens in
                the input and output.
        """
        super().__init__()
        self.trainer = trainer
        self.tokenizer = tokenizer
        self.val_dataset = val_dataset
        self.epoch_count = 0
        self.val_dataset_size = len(self.val_dataset)
        self.executor_batch_size = executor_batch_size
        self.tokenizer_max_length = tokenizer_max_length
        self.sequence_max_length = sequence_max_length

    def on_epoch_end(self, args, state, control, **kwargs):
        """After each  evaluation, this function will be called."""
        _ = (args, state, control, kwargs)
        # Suppress the unused parameters warning.
        self.epoch_count += 1
        logger.info(
            f"Epoch: {self.epoch_count}. Evaluate on { self.val_dataset_size} examples."
        )
        # For multi-GPU training, the training processor will be segmented
        # into multi-threads with data paralyzation, so the validation dataset
        # used in the callback is also segmented.
        model_executor = GenerationModelExecutor(
            model=self.trainer.model,
            tokenizer=self.tokenizer,
            batch_size=self.executor_batch_size,
            tokenizer_max_length=self.tokenizer_max_length,
            sequence_max_length=self.sequence_max_length,
        )
        model_outputs = model_executor.make_prediction(
            self.val_dataset,
            "model_input",
        )
        evaluator = Seq2SeqEvaluator()
        metric_values = evaluator.evaluate_model(
            self.val_dataset,
            "model_output",
            model_outputs,
            encoder_model_name="xlm-roberta-base",
        )
        logger.info(metric_values)



"""An base class for trainers."""
from __future__ import annotations  # noqa FI58

from abc import ABC, abstractmethod
from typing import Any

import datasets
import transformers
from transformers import AutoModel, AutoTokenizer


# pylint: disable=too-few-public-methods
class BaseTrainer(ABC):
    """Train a model with a fixed set of hyperparameters."""

    def __init__(self, pretrained_model_name: str):
        """Initialize a model trainer.

        Args:
            pretrained_model_name: A HuggingFace model name to use for training.
        """
        self.model = AutoModel.from_pretrained(pretrained_model_name)
        self.tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name, padding_side="left"
        )
        self.wandb = None

    @abstractmethod
    def train_model(
        self,
        hyperparameter_choices: dict[str, Any],
        training_datasets: list[datasets.Dataset],
        validation_datasets: list[datasets.Dataset] | None = None,
    ) -> tuple[transformers.PreTrainedModel, transformers.PreTrainedTokenizer]:
        """Train a model with the given hyperparameters and return it."""



"""An interface for generating model outputs."""

from __future__ import annotations  # noqa FI58

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

import datasets
import transformers

from prompt2model.utils import get_formatted_logger

logger = get_formatted_logger("ModelExecutor")


@dataclass(frozen=False)
class ModelOutput:
    """A model output for a single example.

    Attributes:
        prediction: The prediction by the model.
        auxiliary_info: Any other auxiliary information provided by the model.
    """

    prediction: Any
    auxiliary_info: dict[str, Any]


class ModelExecutor(ABC):
    """An interface for automatic model evaluation."""

    def __init__(
        self,
        model: transformers.PreTrainedModel,
        tokenizer: transformers.PreTrainedTokenizer,
        batch_size: int = 10,
        tokenizer_max_length: int = 256,
        sequence_max_length: int = 512,
    ) -> None:
        """Initializes a new instance of ModelExecutor.

        Args:
            model: The model to evaluate.
            tokenizer: The model's associated tokenizer.
            batch_size: The batch size to use when making predictions.
            tokenizer_max_length: The maximum number of tokens that
                tokenizer is allowed to generate.
            sequence_max_length: The maximum number of tokens in
                the input and output.
        """
        self.model = model
        self.tokenizer = tokenizer
        self.batch_size = batch_size
        if self.tokenizer.pad_token is None:
            logger.warning(
                "Trying to init an ModelExecutor's tokenizer without pad_token."
            )
            self.tokenizer.pad_token = self.tokenizer.eos_token
            self.model.config.pad_token_id = self.model.config.eos_token_id
        self.tokenizer_max_length = tokenizer_max_length
        self.sequence_max_length = sequence_max_length
        if self.sequence_max_length is None:
            logger.warning(
                (
                    "The `max_length` in `self.model.generate` will default to "
                    f"`self.model.config.max_length` ({self.model.config.max_length})"
                    " if `sequence_max_length` is `None`."
                )
            )
        if hasattr(self.model.config, "max_position_embeddings"):
            max_embeddings = self.model.config.max_position_embeddings
            if sequence_max_length is not None and sequence_max_length > max_embeddings:
                logger.warning(
                    (
                        f"The sequence_max_length ({sequence_max_length})"
                        f" is larger than the max_position_embeddings ({max_embeddings})."  # noqa: E501
                        f" So the sequence_max_length will be set to {max_embeddings}."  # noqa: E501
                    )
                )
                self.sequence_max_length = max_embeddings

    @abstractmethod
    def make_prediction(
        self,
        test_set: datasets.Dataset,
        input_column: str,
    ) -> list[ModelOutput]:
        """Evaluate a model on a test set.

        Args:
            test_set: The dataset to make predictions on.
            input_column: The dataset column to use as input to the model.

        Returns:
            A list of model outputs, one for each element in the test set.
        """

    @abstractmethod
    def make_single_prediction(self, model_input: str) -> ModelOutput:
        """Make prediction on one example.

        Args:
            model_input: The input string to the model.

        Returns:
            A single model output, useful for exposing a model to a user interface.
        """



"""An interface for prompt parsing."""

from __future__ import annotations  # noqa FI58

from abc import ABC, abstractmethod
from enum import Enum


class TaskType(Enum):
    """High-level taxonomy of possible NLP model outputs."""

    TEXT_GENERATION = 1
    CLASSIFICATION = 2
    SEQUENCE_TAGGING = 3
    SPAN_EXTRACTION = 4


class PromptSpec(ABC):
    """Parse and store structured information about the prompt."""

    task_type: TaskType
    _instruction: str | None
    _examples: str | None

    @abstractmethod
    def parse_from_prompt(self, prompt: str) -> None:
        """Populate this class by parsing a prompt."""

    @property
    def instruction(self) -> str:
        """Return the natural language instruction parsed from the prompt."""
        assert self._instruction is not None
        return self._instruction

    @property
    def examples(self) -> str:
        """Return the natural language examples parsed from the prompt."""
        return self._examples or ""



"""Tools for retrieving dataset metadata from HuggingFace.

Before calling this script, set the HF_USER_ACCESS_TOKEN environment variable.
"""

from __future__ import annotations  # noqa FI58

import argparse
import json

import requests
from huggingface_hub import list_datasets

parser = argparse.ArgumentParser()
parser.add_argument(
    "--dataset-index-file",
    type=str,
    default="huggingface_data/huggingface_datasets/dataset_index.json",
)


def get_fully_supported_dataset_names():
    """Get the list of loadable datasets from HuggingFace."""
    API_URL = "https://datasets-server.huggingface.co/valid"
    response = requests.get(API_URL)
    datasets_list = response.json()
    fully_supported_datasets = datasets_list["viewer"] + datasets_list["preview"]
    return fully_supported_datasets


def construct_search_documents(
    all_dataset_names,
    all_dataset_descriptions,
    fully_supported_dataset_names,
    minimum_description_length=4,
):
    """Select the datasets and corresponding descriptions to store in our index."""
    filtered_dataset_names = []
    nonempty_descriptions = []
    for dataset_name, description in zip(all_dataset_names, all_dataset_descriptions):
        if dataset_name not in fully_supported_dataset_names:
            continue
        if (
            description is not None
            and len(description.split()) > minimum_description_length
        ):
            filtered_dataset_names.append(dataset_name)
            nonempty_descriptions.append(description)
    return filtered_dataset_names, nonempty_descriptions


if __name__ == "__main__":
    args = parser.parse_args()

    fully_supported_dataset_names = get_fully_supported_dataset_names()
    all_datasets = list(list_datasets())
    dataset_names = [dataset.id for dataset in all_datasets]
    dataset_descriptions = [dataset.description for dataset in all_datasets]

    filtered_dataset_names, filtered_descriptions = construct_search_documents(
        dataset_names, dataset_descriptions, fully_supported_dataset_names
    )
    dataset_index = {}
    for name, description in zip(filtered_dataset_names, filtered_descriptions):
        dataset_index[name] = {
            "name": name,
            "description": description,
        }

    json.dump(dataset_index, open(args.dataset_index_file, "w"))



"""A mock dataset retriever for testing purposes."""

from __future__ import annotations  # noqa FI58

import datasets

from prompt2model.dataset_retriever.base import DatasetRetriever
from prompt2model.prompt_parser import PromptSpec


class MockRetriever(DatasetRetriever):
    """A class for retrieving datasets."""

    def __init__(self):
        """Construct a mock dataset retriever."""

    def retrieve_dataset_dict(
        self, prompt_spec: PromptSpec
    ) -> datasets.DatasetDict | None:
        """Return a single empty DatasetDict for testing purposes."""
        _ = prompt_spec  # suppress unused vaiable warning
        mock_dataset = datasets.Dataset.from_dict(
            {"input_col": [""], "output_col": [""]}
        )
        return [
            datasets.DatasetDict(
                {"train": mock_dataset, "val": mock_dataset, "test": mock_dataset}
            )
        ]



"""An interface for prompt parsing."""

from prompt2model.prompt_parser.base import PromptSpec, TaskType


class MockPromptSpec(PromptSpec):
    """Mock the bebavior of PromptSpec."""

    def __init__(
        self, task_type: TaskType, instruction: str = None, examples: str = None
    ):
        """Mock the elements of PromptSpec."""
        self.task_type = task_type
        if instruction is None:
            self._instruction = (
                "Give me some translation from Chinese to English."
                " Input Chinese and output English."
            )
        else:
            self._instruction = instruction
        if examples is None:
            self._examples = (
                "input: '人生苦短，我用 Python', output: 'Life is short, I use Python. '"
                "input: '明天是周末', output: 'Tomorrow is weekend.'"
            )
        else:
            self._examples = examples

    def parse_from_prompt(self, prompt: str) -> None:
        """Don't parse anything."""
        self._instruction = prompt
        return None



"""Tools for testing if two Datasets or DatasetDicts are identical."""


from __future__ import annotations  # noqa FI58

import datasets


def are_datasets_identical(
    dataset1: datasets.Dataset, dataset2: datasets.Dataset
) -> bool:
    """Check if two datasets are identical in terms of instance values and order."""
    if len(dataset1) != len(dataset2):
        return False

    return all(
        instance1 == instance2 for instance1, instance2 in zip(dataset1, dataset2)
    )


def are_dataset_dicts_identical(
    dataset_dict1: datasets.DatasetDict, dataset_dict2: datasets.DatasetDict
) -> bool:
    """Check if two DatasetDict objects are identical."""
    if set(dataset_dict1.keys()) != set(dataset_dict2.keys()):
        return False

    return all(
        are_datasets_identical(dataset_dict1[split_name], dataset_dict2[split_name])
        for split_name in dataset_dict1.keys()
    )



"""Tools for creating a mock search index."""

import pickle

import numpy as np


def create_test_search_index(index_file_name: str) -> None:
    """Utility function to create a test search index.

    This search index represents 3 models, each represented with a hand-written vector.
    Given a query of [0, 0, 1], the 3rd model will be the most similar.

    Args:
        index_file_name: The name of the file to which a pickled index will be written.
    """
    mock_model_encodings = np.array([[0.9, 0, 0], [0, 0.9, 0], [0, 0, 0.9]])
    mock_lookup_indices = [0, 1, 2]
    with open(index_file_name, "wb") as f:
        pickle.dump((mock_model_encodings, mock_lookup_indices), f)



"""Classes for setting random seeds."""

from __future__ import annotations  # noqa FI58

from abc import ABC, abstractmethod


class SeedGenerator(ABC):
    """Select a good model from among a set of hyperparameter choices."""

    @abstractmethod
    def get_seed(self) -> int:
        """Generate a random seed."""


class ConstantSeedGenerator(SeedGenerator):
    """A seed generator that always returns the same seed."""

    def __init__(self, seed: int = 2023):
        """Initialize with a constant seed (by default, 2023)."""
        self.seed = seed

    def get_seed(self) -> int:
        """Return a constant random seed."""
        return self.seed


seed_generator = ConstantSeedGenerator()



"""Import DemoCreator functions."""

from prompt2model.demo_creator.create import create_gradio
from prompt2model.demo_creator.mock import mock_gradio_create

__all__ = (
    "mock_gradio_create",
    "create_gradio",
)



"""A dummy evaluator for testing purposes."""
from __future__ import annotations  # noqa FI58

from typing import Any

import datasets

from prompt2model.model_evaluator.base import ModelEvaluator
from prompt2model.model_executor import ModelOutput
from prompt2model.prompt_parser import PromptSpec


class MockEvaluator(ModelEvaluator):
    """A dummy evaluator that always returns the same metric value."""

    def __init__(self) -> None:
        """Initialize the evaluation setting."""

    def evaluate_model(
        self,
        dataset: datasets.Dataset,
        gt_column: str,
        predictions: list[ModelOutput],
        metrics: list[datasets.Metric] | None = None,
        prompt_spec: PromptSpec | None = None,
    ) -> dict[str, Any]:
        """Return empty metrics dictionary.

        Args:
            dataset: The dataset to evaluate metrics on.
            gt_column: The dataset column to use as ground truth.
            predictions: Corresponding model outputs to evaluate.
            metrics: (Optional) The metrics to use, defaults to using
                chr_f, exact_match, and bert_score.
            prompt_spec: (Optional) A PromptSpec to infer the metrics from.

        Returns:
            An empty dictionary (for testing purposes).
        """
        return {}



"""Import PromptSpec classes."""
from prompt2model.prompt_parser.base import PromptSpec, TaskType
from prompt2model.prompt_parser.instr_parser import OpenAIInstructionParser
from prompt2model.prompt_parser.mock import MockPromptSpec

__all__ = (
    "PromptSpec",
    "TaskType",
    "MockPromptSpec",
    "OpenAIInstructionParser",
)



"""Tools for creating a GPT-2 model with padding tokenizer."""

from collections import namedtuple

from transformers import AutoModelForCausalLM, AutoTokenizer, T5ForConditionalGeneration

ModelAndTokenizer = namedtuple("ModelAndTokenizer", ["model", "tokenizer"])


def create_gpt2_model_and_tokenizer(full_size: bool = False) -> ModelAndTokenizer:
    """Create a GPT2 model with its padding tokenizer for batched input.

    Args:
        full_size: Whether to use the full size of the GPT-2 model. Defaults to False.
            Note that the full size of the GPT-2 model may occupy too much memory
            that lead to out of memory errors.

    Returns:
        gpt2_model_and_tokenizer: A namedtuple with gpt2 model and tokenizer.
    """
    if not full_size:
        gpt2_model = AutoModelForCausalLM.from_pretrained("sshleifer/tiny-gpt2")
        gpt2_tokenizer = AutoTokenizer.from_pretrained("sshleifer/tiny-gpt2")
    else:
        gpt2_model = AutoModelForCausalLM.from_pretrained("gpt2")
        gpt2_tokenizer = AutoTokenizer.from_pretrained("gpt2")
    if gpt2_tokenizer.pad_token is None:
        gpt2_tokenizer.pad_token = gpt2_tokenizer.eos_token
    if gpt2_model.config.pad_token_id is None:
        gpt2_model.config.pad_token_id = gpt2_tokenizer.eos_token_id
    gpt2_model_and_tokenizer = ModelAndTokenizer(gpt2_model, gpt2_tokenizer)
    return gpt2_model_and_tokenizer


def create_t5_model_and_tokenizer(full_size: bool = False) -> ModelAndTokenizer:
    """Create a T5 model with its padding tokenizer for batched input.

    Args:
        full_size: Whether to use the full size of the T5 model. Defaults to False.
            Note that the full size of the T5 model may occupy too much memory
            that lead to out of memory errors.

    Returns:
        t5_model_and_tokenizer: A namedtuple with t5 model and tokenizer.
    """
    if not full_size:
        t5_model = T5ForConditionalGeneration.from_pretrained(
            "google/t5-efficient-tiny"
        )
        t5_tokenizer = AutoTokenizer.from_pretrained("google/t5-efficient-tiny")
    else:
        t5_model = T5ForConditionalGeneration.from_pretrained("t5-small")
        t5_tokenizer = AutoTokenizer.from_pretrained("t5-small")
    t5_model_and_tokenizer = ModelAndTokenizer(t5_model, t5_tokenizer)
    return t5_model_and_tokenizer



"""Import all the model executor classes."""

from prompt2model.model_executor.base import ModelExecutor, ModelOutput
from prompt2model.model_executor.generate import GenerationModelExecutor
from prompt2model.model_executor.mock import MockModelExecutor

__all__ = (
    "ModelExecutor",
    "ModelOutput",
    "MockModelExecutor",
    "GenerationModelExecutor",
)



"""An interface for dataset generation."""

from __future__ import annotations  # noqa FI58

from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path

import datasets

from prompt2model.prompt_parser import PromptSpec


class DatasetSplit(Enum):
    """The split of a dataset."""

    TRAIN = "train"
    VAL = "val"
    TEST = "test"


class DatasetGenerator(ABC):
    """A class for generating datasets from a prompt specification."""

    @abstractmethod
    def generate_dataset_split(
        self,
        prompt_spec: PromptSpec,
        expected_num_examples: int,
        split: DatasetSplit,
    ) -> datasets.Dataset:
        """Generate data for a single named split of data.

        Args:
            prompt_spec: A prompt spec (containing a system description).
            expected_num_examples: Expected number of examples in split.
            split: Name of dataset split to generate.

        Returns:
            A single dataset split.
        """

    def generate_dataset_dict(
        self,
        prompt_spec: PromptSpec,
        expected_num_examples: dict[DatasetSplit, int],
        output_dir: str | None = None,
    ) -> datasets.DatasetDict:
        """Generate full dataset splits (e.g. train/dev/test) from a prompt.

        Args:
            prompt_spec: A prompt specification.
            expected_num_examples: Expected number of
                examples per split (train/val/test).

        Returns:
            A DatasetDict containing train, val, and test splits.
        """
        dataset_dict = datasets.DatasetDict(
            {
                split.value: self.generate_dataset_split(prompt_spec, num, split=split)
                for split, num in expected_num_examples.items()
            }
        )

        if output_dir:
            save_dir = Path(output_dir)
            save_dir.mkdir(parents=True, exist_ok=True)
            dataset_dict.save_to_disk(str(save_dir))

        return dataset_dict



"""Import DatasetProcessor classes."""
from prompt2model.dataset_processor.base import BaseProcessor
from prompt2model.dataset_processor.mock import MockProcessor
from prompt2model.dataset_processor.textualize import TextualizeProcessor

__all__ = ("BaseProcessor", "TextualizeProcessor", "MockProcessor")



"""Script to run the model retriever in isolation."""

from prompt2model.model_retriever import DescriptionModelRetriever
from prompt2model.prompt_parser import MockPromptSpec, TaskType

if __name__ == "__main__":
    prompt_spec = MockPromptSpec(TaskType.TEXT_GENERATION)
    prompt = """Your task is to generate an answer to a natural question. In this task, the input is a string that consists of both a question and a context passage. The context is a descriptive passage related to the question and contains the answer. And the question can range from Math, Cultural, Social, Geometry, Biology, History, Sports, Technology, Science, and so on."""  # noqa E501
    prompt_spec._instruction = prompt
    retriever = DescriptionModelRetriever(
        search_index_path="huggingface_data/huggingface_models/bm25_search_index.pkl",
        model_descriptions_index_path="huggingface_data/huggingface_models/model_info/",
        use_bm25=True,
        use_HyDE=True,
    )
    top_model_name = retriever.retrieve(prompt_spec)



"""Import Tevatron utility functions."""
from prompt2model.utils.tevatron_utils.encode import encode_text
from prompt2model.utils.tevatron_utils.retrieve import retrieve_objects

__all__ = ("encode_text", "retrieve_objects")



"""A mock dataset processor for testing purposes."""

import datasets

from prompt2model.dataset_processor.base import BaseProcessor


class MockProcessor(BaseProcessor):
    """A class for retrieving datasets."""

    def process_dataset_dict(
        self, instruction: str, dataset_dicts: list[datasets.DatasetDict]
    ) -> list[datasets.DatasetDict]:
        """A mock function to post-process a list of DatasetDicts.

        Args:
            instruction: The instruction used as a prefix to explain the task.
            dataset_dicts: A list of DatasetDicts (generated or retrieved).

        Returns:
            A list of DatasetDicts, all examples are converted into text2text fashion.
        """
        _ = instruction
        return dataset_dicts

    def post_process_example(example: dict, instruction: str, task_id: int) -> dict:
        """A mock function that modifies a given example dictionary.

        Args:
            example: A dictionary representing an example.
            instruction: The instruction used as a prefix to explain the task.
            task_id: A tag marking which dataset (from dataset_dicts) this example
                comes from. Used for multi-task training.
            has_encoder: Whether the retrieved model has an encoder.
            dataset_split: The split of the example, i.e. train/val/test.
            eos_token: The end-of-sentence token of the tokenizer.

        Returns:
            A dictionary with `model_input` as the input to models
            and `model_output` as the expected output of models.
        """
        _ = instruction, task_id
        example["model_input"] = example["input_col"]
        example["model_output"] = example["output_col"]
        return example
